

"""
 Extract base level metrics
"""
# standard imports
import math

# third-party imports
import numpy as np
import opendssdirect as dss
import pandas as pd

# internal imports
from emerge.network.asset_metrics import networkx_from_opendss_model


def get_allbus_voltage_pu(dss_instance):
    return dss_instance.Circuit.AllBusMagPu()

def get_voltage_distribution(voltage_array:list, 
    bins:list = [0.7 + i*0.05 for i in range(12)]):

    # Remove zeros
    voltage_array = np.array(voltage_array)
    voltage_array = voltage_array[voltage_array !=0]

    hist, _ = np.histogram(voltage_array, bins)
    pct_hist = list(np.array(hist) *100/ len(voltage_array))
    
    return {
        'Voltage bins': bins[:-1],
        'Percentage nodes':  pct_hist
    }

def get_voltage_by_distance(dss_instance):


    mapper = {1: 'Phase A', 2: 'Phase B', 3: 'Phase C'}
    voltage_by_distance = {'Voltage (pu)': [], 'Distance from substation (km)': [], 'Phase': []}
    for phase in range(1,4):
        pu_voltages = dss_instance.Circuit.AllNodeVmagPUByPhase(phase)
        node_distances = dss_instance.Circuit.AllNodeDistancesByPhase(phase)

        voltage_by_distance['Voltage (pu)'].extend(pu_voltages)
        voltage_by_distance['Phase'].extend([mapper[phase]]*len(pu_voltages))
        voltage_by_distance['Distance from substation (km)'].extend(node_distances)

    return voltage_by_distance

def get_voltage_by_lat_lon(dss_instance):

    network = networkx_from_opendss_model(dss_instance)
    node_data = {node[0]: node[1] for node in network.nodes.data()}
    all_bus_voltage = dss_instance.Circuit.AllBusMagPu()
    all_bus_names = dss_instance.Circuit.AllBusNames()

    voltage_by_lat_lon = {
        'longitudes': [],
        'latitudes': [],
        'voltage (pu)': []
    }

    for bus, voltage_pu in zip(all_bus_names, all_bus_voltage):
        voltage_by_lat_lon['latitudes'].append(node_data[bus]['pos'][1])
        voltage_by_lat_lon['longitudes'].append(node_data[bus]['pos'][0])
        voltage_by_lat_lon['voltage (pu)'].append(voltage_pu)
    
    return voltage_by_lat_lon

def get_voltage_dataframe(dss_instance: dss):
    """ Function to retrieve voltage dataframe for all buses. 
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect
    """

    all_bus_voltage = dss_instance.Circuit.AllBusMagPu()
    all_bus_names = dss_instance.Circuit.AllBusNames()

    voltage_df = {"busname": [], "voltage(pu)": []}

    for bus, voltage_pu in zip(all_bus_names, all_bus_voltage):
        voltage_df['busname'].append(bus)
        voltage_df['voltage(pu)'].append(voltage_pu)
    
    return pd.DataFrame(voltage_df).set_index("busname")


def get_lineloading_dataframe(dss_instance: dss):
    """ Function to retrieve line loading dataframe for all line segments. 
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect
    """
    line_loading_df = {"linename": [], "loading(pu)": []}

    dss_instance.Circuit.SetActiveClass("Line")
    flag = dss_instance.ActiveClass.First()

    while flag:
        line_name = dss_instance.CktElement.Name().lower()
        n_phases = dss_instance.CktElement.NumPhases()
        line_limit = dss_instance.CktElement.NormalAmps()
        currents = dss_instance.CktElement.CurrentsMagAng()[:2 * n_phases]
        line_current = currents[::2]
        ldg = max(line_current) / max(float(line_limit), 1)
        line_loading_df['linename'].append(line_name)
        line_loading_df['loading(pu)'].append(ldg)
        flag = dss_instance.ActiveClass.Next()
    
    return pd.DataFrame(line_loading_df).set_index("linename")

def get_pv_power_dataframe(dss_instance: dss):
    """ Function to retrieve pv power dataframe.
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect
    """
    pv_power_df = {"pvname": [], "active_power": [], "reactive_power": []}

    flag = dss_instance.PVsystems.First()
    while flag>0:
        pv_name = dss_instance.PVsystems.Name().lower()
        pv_powers = dss_instance.CktElement.Powers()
        
        active_power = -sum(pv_powers[::2]) 
        reactive_power = -sum(pv_powers[1::2]) 
        pv_power_df['pvname'].append(pv_name)
        pv_power_df['active_power'].append(active_power)
        pv_power_df['reactive_power'].append(reactive_power)
        
        flag = dss_instance.PVsystems.Next()
    
    return pd.DataFrame(pv_power_df).set_index("pvname")



def get_transloading_dataframe(dss_instance: dss):
    """ Function to retrieve transformer loading dataframe for all transformers. 
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect
    """
    trans_loading_df = {"transformername": [], "loading(pu)": []}

    dss_instance.Circuit.SetActiveClass("Transformer")
    flag = dss_instance.ActiveClass.First()

    while flag:
        xfmr_name = dss_instance.CktElement.Name().lower()
        n_phases = dss_instance.CktElement.NumPhases()
        hs_kv = float(dss_instance.Properties.Value('kVs').split('[')[1].split(',')[0])
        kva = float(dss_instance.Properties.Value('kVA'))
        if n_phases > 1:
            xfmr_limit = kva / (hs_kv * math.sqrt(3))
        else:
            xfmr_limit = kva / (hs_kv)
        currents = dss_instance.CktElement.CurrentsMagAng()[:2 * n_phases]
        xfmr_current = currents[::2]
        ldg = max(xfmr_current) / float(xfmr_limit)
        
        trans_loading_df['transformername'].append(xfmr_name)
        trans_loading_df['loading(pu)'].append(ldg)
        flag = dss_instance.ActiveClass.Next()
    
    return pd.DataFrame(trans_loading_df).set_index("transformername")




""" OpenDSS simulator service for performing power flow."""

# standard imports
from pathlib import Path
import logging
from abc import ABC, abstractmethod

# third-party imports
import opendssdirect as dss

# internal imports
from emerge.utils.util import validate_path

logger = logging.getLogger(__name__)


class AbstractSimulator(ABC):

    @abstractmethod
    def set_stepsize(self, step_in_min):
        pass

    @abstractmethod
    def solve(self):
        pass

    @abstractmethod
    def set_simulation_time(self, sim_time, profile_start_time):
        pass 


class OpenDSSSimulator(AbstractSimulator):

    def __init__(self, 
            path_to_master_dss_file,
        ):

        validate_path(path_to_master_dss_file, check_for_dir=False, \
            check_for_file=True, file_types=['.dss'])
        
        self.case_file = Path(path_to_master_dss_file)
        self.dss_instance = dss

        self.dss_instance.run_command("Clear")
        self.dss_instance.Basic.ClearAll()
        self.execute_dss_command(f'Redirect {self.case_file}')

    def execute_dss_command(self, dss_command: str):
        
        error = self.dss_instance.run_command(dss_command)
        if error:
            logger.error(f"Error executing command {dss_command} >> {error}")
            raise Exception(f"Error executing command {dss_command} >> {error}")
        logger.info(f"Sucessfully executed the command, {dss_command}")


    def set_frequency(self, frequency):
        self.execute_dss_command(f"Set DefaultBaseFrequency={frequency}")
        

    def set_mode(self, mode):
        ## Use it wisely if you need to e.g. mode =2 means QSTS powerflow
        self.dss_instance.Solution.Mode(mode)

    def set_max_iteration(self, max_iterations):
        self.dss_instance.Solution.MaxControlIterations(max_iterations)

    def set_stepsize(self, step_in_min):
        self.dss_instance.Solution.StepSizeMin(step_in_min)

    def solve(self):

        self.dss_instance.Solution.Number(1)
        self.dss_instance.Solution.Solve()
        return self.dss_instance.Solution.Converged()

    def set_simulation_time(self, sim_time, profile_start_time):

        time_diff = sim_time - profile_start_time
        total_seconds = time_diff.total_seconds()
        hours = int(total_seconds/3600)
        seconds = total_seconds - hours*3600
        self.dss_instance.Solution.Hour(hours)
        self.dss_instance.Solution.Seconds(seconds)




""" This module contains test function for testing the
feature included in `powerflow_metrics_powerflow` module."""

from pathlib import Path

import pandas as pd

from emerge.simulator import powerflow_results
from emerge.simulator import opendss

def test_pv_power_dataframe():
    """ Test function for `pv_power_dataframe` 
    utility function."""

    root_path = Path(__file__).absolute().parents[1] 
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    pv_df = powerflow_results.get_pv_power_dataframe(
        simulator.dss_instance
    )

    assert isinstance(pv_df, pd.DataFrame)

def test_get_voltage_by_dataframe():
    """ Test function for `get_voltage_by_dataframe` 
    utility function."""

    root_path = Path(__file__).absolute().parents[1] 
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    voltage_df = powerflow_results.get_voltage_dataframe(
        simulator.dss_instance
    )

    assert isinstance(voltage_df, pd.DataFrame)

def test_get_lineloading_dataframe():
    """ Test function for `get_lineloading_dataframe` 
    utility function."""

    root_path = Path(__file__).absolute().parents[1] 
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    line_loading_df = powerflow_results.get_lineloading_dataframe(
        simulator.dss_instance
    )

    assert isinstance(line_loading_df, pd.DataFrame)

def test_get_transloading_dataframe():
    """ Test function for `get_transloading_dataframe` 
    utility function."""

    root_path = Path(__file__).absolute().parents[1] 
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    xfmr_loading_df = powerflow_results.get_transloading_dataframe(
        simulator.dss_instance
    )

    assert isinstance(xfmr_loading_df, pd.DataFrame)





"""
 Extract base level metrics
"""
# standard imports

# third-party imports
import networkx as nx

# internal imports
from emerge.network.data_model import (
    AssetMetrics,
    FeederMetrics,
    LoadAssetMetrics,
    PVAssetMetrics,
    CapacitorAssetMetrics,
    RegulatorsAssetMetrics,
    TransformersAssetMetrics
)

def opendss_load_metrics_extractor(dss_instance):

    flag = dss_instance.Loads.First()
    kws, kvars = [], []
    while flag:
        kws.append(dss_instance.Loads.kW())
        kvars.append(dss_instance.Loads.kvar())
        flag = dss_instance.Loads.Next()

    return LoadAssetMetrics(
        total_count=dss_instance.Loads.Count(),
        max_kw_capacity=round(max(kws), 3) if kws else 0,
        min_kw_capacity=round(min(kws), 3) if kws else 0,
        min_kvar_capacity=round(min(kvars), 3) if kvars else 0,
        max_kvar_capacity=round(max(kvars), 3) if kvars else 0,
        total_kw_capacity=round(sum(kws), 3) if kws else 0,
        total_kvar_capacity=round(sum(kvars), 3) if kvars else 0
    )



def opendss_pv_metrics_extractor(dss_instance):

    flag = dss_instance.PVsystems.First()
    kws = []
    while flag:
        kws.append(dss_instance.PVsystems.Pmpp())
        flag = dss_instance.PVsystems.Next()

    return PVAssetMetrics(
        total_count=dss_instance.PVsystems.Count(),
        max_kw_capacity=round(max(kws), 3) if kws else 0,
        min_kw_capacity=round(min(kws), 3) if kws else 0,
        total_kw_capacity=round(sum(kws), 3) if kws else 0
    )

def opendss_capacitors_metrics_extractor(dss_instance):

    flag = dss_instance.Capacitors.First()
    kvars = []
    while flag:
        kvars.append(dss_instance.Capacitors.kvar())
        flag = dss_instance.PVsystems.Next()

    return CapacitorAssetMetrics(
        total_count=dss_instance.Capacitors.Count(),
        max_kvar_capacity=round(max(kvars), 3) if kvars else 0,
        min_kvar_capacity=round(min(kvars), 3) if kvars else 0,
        total_kvar_capacity=round(sum(kvars), 3) if kvars else 0
    )

def opendss_regulators_metrics_extractor(dss_instance):

    return RegulatorsAssetMetrics(
        total_count=dss_instance.RegControls.Count()
    )

def opendss_transformers_metrics_extractor(dss_instance):

    flag = dss_instance.Transformers.First()
    kvas = []
    while flag:
        kvas.append(dss_instance.Transformers.kVA())
        flag = dss_instance.Transformers.Next()

    return TransformersAssetMetrics(
        total_count=dss_instance.Transformers.Count(),
        max_kva_capacity=round(max(kvas), 3) if kvas else 0,
        min_kva_capacity=round(min(kvas), 3) if kvas else 0,
        total_kva_capacity=round(sum(kvas), 3) if kvas else 0
    )

def networkx_from_opendss_model(dss_instance):

    """ Let's create a networkx representation of the model"""
    network = nx.Graph()

    """ Let's add all buses """
    for bus in dss_instance.Circuit.AllBusNames():
        dss_instance.Circuit.SetActiveBus(bus)
        network.add_node(bus, pos=(dss_instance.Bus.X(), \
            dss_instance.Bus.Y()))

    UNIT_MAPPER = {
        0: 0,
        1: 1.60934,
        2: 0.3048,
        3: 1,
        4: 0.001,
        5: 0.0003048,
        6: 0.0000254,
        7: 0.00001
    }

    """ Let's all line elements to the network """

    for edge_name in ['Line', 'Transformer']:
        dss_instance.Circuit.SetActiveClass(edge_name)
        flag =dss_instance.ActiveClass.First()
        while flag > 0:
            name = dss_instance.CktElement.Name().lower()
            buses = dss_instance.CktElement.BusNames()
            edge_length = 0 if edge_name == 'Transformer' else \
                UNIT_MAPPER[dss_instance.Lines.Units()]*dss_instance.Lines.Length()

            network.add_edge(buses[0].split('.')[0], buses[1].split('.')[0], name = name, \
                length=edge_length)
            flag = dss_instance.ActiveClass.Next()

    return network

def get_all_kv_levels(dss_instance):
    
    flag = dss_instance.Transformers.First()
    kvs = []
    while flag:
        for wdg in range(1,dss_instance.Transformers.NumWindings()+1):
            dss_instance.Transformers.Wdg(wdg)
            kvs.append(dss_instance.Transformers.kV())
        flag = dss_instance.Transformers.Next()
    return list(set(kvs))


def opendss_feeder_metrics_extractor(dss_instance):

    df = dss_instance.utils.class_to_dataframe('vsource').to_dict()
    source_bus = df['bus1']['vsource.source'].split('.')[0]

    """ Get networkx representation of opendss model """
    network = networkx_from_opendss_model(dss_instance)


    """ Let's create dfs tree """
    dfs_tree = nx.dfs_tree(network, source=source_bus)

    for edge in dfs_tree.edges():
        edge_data = network[edge[0]][edge[1]]
        dfs_tree[edge[0]][edge[1]]['length'] = edge_data['length']
        
    total_feeder_length = nx.dag_longest_path_length(dfs_tree, weight="length", default_weight=0)
    longest_path_length = nx.dag_longest_path(dfs_tree, weight="length", default_weight=0)
    
    """ Get all the kvs """
    all_kvs = get_all_kv_levels(dss_instance)
    all_kvs.sort()
    secondary_kv, primary_kv = all_kvs[0], all_kvs[1]

    transformer_buses = []

    flag = dss_instance.Transformers.First()
    
    while flag:
        buses = dss_instance.CktElement.BusNames()
       
        kvs = []
        for wdg in range(1,dss_instance.Transformers.NumWindings()+1):
            dss_instance.Transformers.Wdg(wdg)
            kvs.append(dss_instance.Transformers.kV())
        if secondary_kv in kvs:
            transformer_buses.append(buses[0].split('.')[0])
        flag = dss_instance.Transformers.Next()

    secondary_lengths = {}
    
    for bus in transformer_buses:
        dfs_tree_ = nx.dfs_tree(dfs_tree, source=bus)
        for edge in dfs_tree_.edges():
            edge_data = network[edge[0]][edge[1]]
            dfs_tree_[edge[0]][edge[1]]['length'] = edge_data['length']
       
        secondary_length = nx.dag_longest_path_length(dfs_tree_, weight="length", default_weight=0)
        secondary_lengths[bus] = secondary_length

    max_primary_feeder_length = total_feeder_length
    for bus, length in secondary_lengths.items():
        if bus in longest_path_length:
            max_primary_feeder_length = total_feeder_length - length

    secondary_lengths_list = list(secondary_lengths.values())
    
   
    return FeederMetrics(
        total_feeder_length_km= round(total_feeder_length, 3),
        max_primary_feeder_length_km= round(max_primary_feeder_length, 3),
        max_secondary_feeder_length_km=round(max(secondary_lengths_list),3) \
            if secondary_lengths_list else 0,
        min_secondary_feeder_length_km=round(min(secondary_lengths_list), 3) \
            if secondary_lengths_list else 0,
        secondary_kv_level=secondary_kv,
        primary_kv_level=primary_kv,
        total_buses=dss_instance.Circuit.NumBuses(),
        total_line_sections=dss_instance.Lines.Count()
    )

def aggregate_asset_metrics(dss_instance):

    loads = opendss_load_metrics_extractor(dss_instance)
    pvs = opendss_pv_metrics_extractor(dss_instance)
    capacitors = opendss_capacitors_metrics_extractor(dss_instance)
    regs = opendss_regulators_metrics_extractor(dss_instance)
    transformers = opendss_transformers_metrics_extractor(dss_instance)
    feeder = opendss_feeder_metrics_extractor(dss_instance)

    return AssetMetrics(
        loads = loads,
        pvs = pvs,
        capacitors=capacitors,
        regulators=regs,
        transformers=transformers,
        lines=feeder
    )


def get_coordinates(dss_instance):

    network = networkx_from_opendss_model(dss_instance)
    coordinates = {
        'nodes': {
            "latitudes": [],
            "longitudes": [],
        },
        "edges": {
            "latitudes": [],
            "longitudes": []
        }
    }

    node_data = {node[0]: node[1] for node in network.nodes.data()}
    for edge in network.edges():
        try:
            from_node = node_data[edge[0]]['pos']
            to_node = node_data[edge[1]]['pos']
            coordinates['nodes']['latitudes'].extend([from_node[1], to_node[1]])
            coordinates['nodes']['longitudes'].extend([from_node[0], to_node[0]])
            coordinates['edges']['latitudes'].extend([from_node[1], to_node[1], None])
            coordinates['edges']['longitudes'].extend([from_node[0], to_node[0], None])
        except Exception as e:
            print(f"{edge} can't get coord info {network.get_edge_data(*edge)}")
        
    return coordinates





""" Module for managing computation of node level metrics. """

# third-party imports
import opendssdirect as dss
import pandas as pd

# internal imports
from emerge.metrics import observer
from emerge.metrics import data_model
from emerge.simulator import powerflow_results
from emerge.utils import dss_util


class TLRI(observer.MetricObserver):
    """Class for managing the computation of Transformer loading risk index (TLRI).

    Attributes:
        loading_threshold (float): Line loading threshol
        tlri_metric (float): TLRI metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        xfmr_downward_customers (pd.DataFrame): Dataframe containing mapping
            between transformer and number of downward customers
    """

    def __init__(self, loading_threshold: float = 1.0):
        """Constructor for LLRI metric.

        Args:
            loading_limit (float): Per unit loading limit
                used for computing SARDI_line metric.
        """

        self.loading_limit = data_model.ThermalLoadingLimit(
            threshold=loading_threshold
        )

        self.tlri_metric = pd.DataFrame()
        self.counter = 0

    def _get_initial_dataset(self, dss_instance: dss):
        """Get initial dataset for computing the metric."""

        self.xfmr_downward_customers = dss_util.get_transformer_customers(
            dss_instance
        )

    def compute(self, dss_instance: dss):
        """Refer to base class for more details."""

        # Get voltage dataframe and load bus mapper
        xmfr_loading_df = powerflow_results.get_transloading_dataframe(
            dss_instance
        )

        if not self.counter:
            self._get_initial_dataset(dss_instance)

        ol_flags = xmfr_loading_df["loading(pu)"] > self.loading_limit.threshold
        ol_gamma = xmfr_loading_df[ol_flags] - self.loading_limit.threshold
        ol_gamma = ol_gamma.rename(columns={"loading(pu)": "ol_llri"})

        merged_df = pd.merge(
            self.xfmr_downward_customers,
            ol_gamma,
            how="left",
            left_index=True,
            right_index=True,
        ).fillna(0)

        merged_df["metric"] = merged_df["customers"] * (merged_df["ol_llri"])

        if self.tlri_metric.empty:
            self.tlri_metric = pd.DataFrame(merged_df["metric"])
        else:
            self.tlri_metric = pd.DataFrame(
                self.tlri_metric["metric"] + merged_df["metric"]
            )

        self.counter += 1

    def get_metric(self):
        """Refer to base class for more details."""
        return {
            xfmrname: metric / self.counter if self.counter else 0
            for xfmrname, metric in self.tlri_metric.to_dict()
            .get("metric", {})
            .items()
        }


class LLRI(observer.MetricObserver):
    """Class for managing the computation of Line loading risk index (NVRI).

    Attributes:
        loading_threshold (float): Line loading threshol
        llri_metric (float): LLRI metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        line_downward_customers (pd.DataFrame): Dataframe containing mapping
            between line segment and number of downward customers
    """

    def __init__(self, loading_threshold: float = 1.0):
        """Constructor for LLRI metric.

        Args:
            loading_limit (float): Per unit loading limit
                used for computing SARDI_line metric.
        """

        self.loading_limit = data_model.ThermalLoadingLimit(
            threshold=loading_threshold
        )

        self.llri_metric = pd.DataFrame()
        self.counter = 0

    def _get_initial_dataset(self, dss_instance: dss):
        """Get initial dataset for computing the metric."""

        self.line_downward_customers = dss_util.get_line_customers(dss_instance)

    def compute(self, dss_instance: dss):
        """Refer to base class for more details."""

        # Get voltage dataframe and load bus mapper
        line_loading_df = powerflow_results.get_lineloading_dataframe(
            dss_instance
        )

        if not self.counter:
            self._get_initial_dataset(dss_instance)

        ol_flags = line_loading_df["loading(pu)"] > self.loading_limit.threshold
        ol_gamma = line_loading_df[ol_flags] - self.loading_limit.threshold
        ol_gamma = ol_gamma.rename(columns={"loading(pu)": "ol_llri"})

        merged_df = pd.merge(
            self.line_downward_customers,
            ol_gamma,
            how="left",
            left_index=True,
            right_index=True,
        ).fillna(0)

        merged_df["metric"] = merged_df["customers"] * (merged_df["ol_llri"])

        if self.llri_metric.empty:
            self.llri_metric = pd.DataFrame(merged_df["metric"])
        else:
            self.llri_metric = pd.DataFrame(
                self.llri_metric["metric"] + merged_df["metric"]
            )

        self.counter += 1

    def get_metric(self):
        """Refer to base class for more details."""
        return {
            linename: metric / self.counter if self.counter else 0
            for linename, metric in self.llri_metric.to_dict()
            .get("metric", {})
            .items()
        }


class NVRI(observer.MetricObserver):
    """Class for managing the computation of Nodal voltage risk index (NVRI).

    Attributes:
        upper_threshold (float): Voltage upper threshold
        lower_threshold (float): Voltage lower threshold
        nvri_metric (float): NVRI metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        bus_load_flag_df (pd.DataFrame): Dataframe containing mapping
            between bus and load flag
    """

    def __init__(
        self, upper_threshold: float = 1.05, lower_threshold: float = 0.95
    ):
        """Constructor for SARDI_voltage metric.

        Args:
            upper_threshold (float): Voltage upper threshold
            lower_threshold (float): Voltage lower threshold
        """

        self.voltage_limit = data_model.VoltageViolationLimit(
            overvoltage_threshold=upper_threshold,
            undervoltage_threshold=lower_threshold,
        )

        self.nvri_metric = pd.DataFrame()
        self.counter = 0

    def _get_initial_dataset(self, dss_instance: dss):
        """Get initial dataset for computing the metric."""

        self.bus_load_flag_df = dss_util.get_bus_load_flag(dss_instance)

    def compute(self, dss_instance: dss):
        """Refer to base class for more details."""

        # Get voltage dataframe and load bus mapper
        voltage_df = powerflow_results.get_voltage_dataframe(
            dss_instance
        )

        if not self.counter:
            self._get_initial_dataset(dss_instance)

        ov_flags = (
            voltage_df["voltage(pu)"] > self.voltage_limit.overvoltage_threshold
        )
        ov_gamma = (
            voltage_df[ov_flags] - self.voltage_limit.overvoltage_threshold
        )
        ov_gamma = ov_gamma.rename(columns={"voltage(pu)": "ov_nvri"})

        uv_flags = (
            voltage_df["voltage(pu)"]
            < self.voltage_limit.undervoltage_threshold
        )
        uv_gamma = (
            self.voltage_limit.undervoltage_threshold - voltage_df[uv_flags]
        )
        uv_gamma = uv_gamma.rename(columns={"voltage(pu)": "uv_nvri"})

        merged_df_ = pd.merge(
            self.bus_load_flag_df,
            ov_gamma,
            how="left",
            left_index=True,
            right_index=True,
        ).fillna(0)
        merged_df = pd.merge(
            merged_df_, uv_gamma, how="left", left_index=True, right_index=True
        ).fillna(0)

        merged_df["metric"] = merged_df["is_load"] * (
            merged_df["ov_nvri"] + merged_df["uv_nvri"]
        )

        ## Deal with duplicate indexes
        merged_df = merged_df[~merged_df.index.duplicated(keep="first")]

        if self.nvri_metric.empty:
            self.nvri_metric = pd.DataFrame(merged_df["metric"])
        else:
            self.nvri_metric = pd.DataFrame(
                self.nvri_metric["metric"] + merged_df["metric"]
            )

        self.counter += 1

    def get_metric(self):
        """Refer to base class for more details."""
        return {
            busname: metric / self.counter if self.counter else 0
            for busname, metric in self.nvri_metric.to_dict()
            .get("metric", {})
            .items()
        }



""" Module for handling data model used in scenario generation. """


from typing import List, Optional, Dict
from enum import Enum

# pylint: disable=no-name-in-module
from pydantic import BaseModel, Field, model_validator
from pydantic_core import PydanticCustomError
from typing_extensions import Annotated, Literal

# pylint: disable=too-few-public-methods

class CustomerModel(BaseModel):
    """Interface for representing customer used in pv scenario development."""

    name: str
    kw: float
    distance: float
    cust_type: Optional[str] = None

class DERType(str, Enum):
    """Available der enum types."""
    # pylint: disable=invalid-name
    load = 'load'
    solar = 'solar'

class CapacityStrategyEnum(str, Enum):
    """Available strategies for sizing solar systems."""

    # pylint: disable=invalid-name
    solar_energy_based = "solar_energy_based"
    peak_multiplier = "peak_multiplier"
    fixed_sizing = "fixed_sizing"


class SelectionStrategyEnum(str, Enum):
    """Available strategies for selecting loads to have pv system."""

    # pylint: disable=invalid-name
    random_allocation = "random_allocation"
    far_allocation = "far_allocation"
    near_allocation = "near_allocation"


class EnergyBasedSolarSizingStrategyInput(BaseModel):
    """Input model for default solar sizing strategy."""

    capacity_factor: Annotated[float, Field(gt=0, le=1)] = 0.3
    load_factor: Annotated[float, Field(gt=0, le=1)] = 0.33
    max_pct_production: Annotated[float, Field(gt=0, le=500)] = 100
    profile: str


class SizeWithProbabilityModel(BaseModel):
    """ Use this model if you want to pick sizes with probabilities. """

    sizes: List[float] | float
    probabilites: List[Annotated[float, Field(ge=0, le=1)]] | float
    profile: List[str] | str
    

    @model_validator(mode="after")
    def validate_lengths_are_same(self) -> 'SizeWithProbabilityModel':
        if isinstance(self.sizes, list) or isinstance(self.probabilites, list) \
            or isinstance(self.profile, list):
            
            try:
                if len(self.sizes) == len(self.probabilites) == len(self.profile):
                    return self 
                raise ValueError("Length must match for sizes,  probabilities and profile names. ")
            except Exception as error:
                raise PydanticCustomError(
                    'mixed_type',
                    f"Error validating fields : {error}"
                )
        return self

class _DERScenarioInput(BaseModel):
    """ Base input model for der scenarios. """
    sizing_strategy: CapacityStrategyEnum
    der_type: DERType
    der_tag: str = ''
    energy_sizing_input:  Dict[str,EnergyBasedSolarSizingStrategyInput] | EnergyBasedSolarSizingStrategyInput | None = None
    peakmult_sizing_input: Dict[str,SizeWithProbabilityModel] | SizeWithProbabilityModel | None = None
    fixed_sizing_input: Dict[str, SizeWithProbabilityModel] | SizeWithProbabilityModel | None = None


class DERScenarioInput(_DERScenarioInput):
    """ Input model for der scenarios. """
    selection_strategy: SelectionStrategyEnum
    other_ders: List[_DERScenarioInput]


class BasicDERModel(BaseModel):
    """Basic DER model used in solar scenario development."""

    name: str
    kw: Annotated[float, Field(ge=0)]
    customer: CustomerModel
    der_type: DERType
    der_tag: str = ''
    profile: str

class DistDERScenarioModel(BaseModel):
    """Model for storing solars in a given scenario."""

    name: str
    sample_id: int
    penetration: float
    ders: List[BasicDERModel]


class LoadMetadataModel(BaseModel):
    """Interface for representing OpenDSS load metadata."""

    name: str
    bus: str
    num_phase: Annotated[int, Field(ge=1, le=3)]
    kv: Annotated[float, Field(gt=0)]
    yearly: Optional[str] = None



class DERScenarioInputModel(_DERScenarioInput):
    """ Interface for der scenario input model. """
    file_name: str
    selection_strategy: SelectionStrategyEnum
    other_ders: List[_DERScenarioInput]


class ScenarioBaseConfig(BaseModel):
    """ Interface for basic settings in defining scenario."""
    pct_resolution: Annotated[float, Field(gt=0, le=100)] = Field(
        description="Percentage resolution or step resolution"
    )
    num_of_penetration: Annotated[int, Field(gt=0)]
    max_num_of_samples: Annotated[int, Field(ge=1)] = 1

class DERScenarioConfigModel(ScenarioBaseConfig):
    """CLI interface model for generating solar scenarios."""

    master_file: str
    output_folder: str
    opendss_attr: Literal['yearly', 'class'] = 'class'
    der_scenario: List[DERScenarioInputModel]



"""Module for managing pv sizing strategies when creating
solar scenarios. """

import abc
from typing import Union, Dict, Optional, Tuple
from emerge.scenarios.data_model import (
    EnergyBasedSolarSizingStrategyInput,
    CustomerModel,
    CapacityStrategyEnum,
    SizeWithProbabilityModel
)
import numpy as np

def _get_value_from_proability(config: Optional[SizeWithProbabilityModel]):
    """ Returns specific value from size with probability model. """

    if config is None:
        raise ValueError(f"Input can not be null >> {config}")
        
    return config.sizes if isinstance(config.sizes, float) \
        else np.random.choice(config.sizes, p=config.probabilites)


# pylint: disable=too-few-public-methods
class SizingStrategy(abc.ABC):
    """Abstract class for sizing strategy."""

    @abc.abstractmethod
    def return_kw_and_profile(self, customer: CustomerModel) -> Optional[Tuple[float, str]]:
        """Abstract method for returning size."""

class EnergyBasedSolarSizingStrategy(SizingStrategy):
    """Default strategy is to size solar based on
    load factor, solar capacity factor and percentage energy
    to be replaced by solar annually."""

    def __init__(
        self,
        config: Union[
            EnergyBasedSolarSizingStrategyInput, Dict[str, EnergyBasedSolarSizingStrategyInput]
        ],
    ):
        self.config = config
       
    def return_kw_and_profile(self, customer: CustomerModel) -> Optional[Tuple[float, str]]:
        
        def _return_size(load_kw: float, _config: EnergyBasedSolarSizingStrategyInput):
            return round(
                (load_kw * _config.load_factor * _config.max_pct_production)
                / (100 * _config.capacity_factor),
                3,
            )

        if isinstance(self.config, dict):
            if customer.cust_type not in self.config:
                print(
                    f"{customer.cust_type} missing from config data {self.config}"
                )
                return (None, None)
            (_return_size(customer.kw, self.config.get(customer.cust_type)), 
             self.config.get(customer.cust_type).profile)
        (_return_size(customer.kw, self.config), self.config.profile)


class PeakMultiplierSizingStrategy(SizingStrategy):
    """Peak multiplier strategy multiplier peak kw with user defined
    multiplier to get the DER size. e.g. ReOpt suggested solar multiplier."""

    def __init__(
        self,
        config: Union[
            SizeWithProbabilityModel,
            Dict[str, SizeWithProbabilityModel],
        ],
    ):
        self.config = config

 
    def return_kw_and_profile(
        self,
        customer: CustomerModel,
    ) -> Optional[Tuple[float, str]]:
        
        def _get_kw_and_profile(config: Optional[SizeWithProbabilityModel]):
            
            multiplier = _get_value_from_proability(config)
            profile = dict(zip(config.sizes, config.profile)).get(multiplier) \
                if isinstance(config.profile, list) else config.profile
            
            return (round(
                customer.kw * multiplier , 3
            ), profile)
        
        if isinstance(self.config, dict):
            if customer.cust_type not in self.config:
                print(
                    f"{customer.cust_type} missing from config data {self.config}"
                )
                return (None, None)
            return _get_kw_and_profile(self.config.get(customer.cust_type))
        
        return _get_kw_and_profile(self.config)
    
class FixedSizingStrategy(SizingStrategy):
    """Fixed sizing strategy uses fixed kw with user defined
    value to get the DER size. e.g. Level 1 EV charger"""

    def __init__(
        self,
        config: Union[
            SizeWithProbabilityModel,
            Dict[str, SizeWithProbabilityModel],
        ],
    ):
        self.config = config


    def return_kw_and_profile(
        self,
        customer: CustomerModel,
    ) -> Optional[Tuple[float, str]]:
        
        def _get_kw_and_profile(config: Optional[SizeWithProbabilityModel]):
            
            fixed_value = _get_value_from_proability(config)
            profile = dict(zip(config.sizes, config.profile)).get(fixed_value) \
                if isinstance(config.profile, list) else config.profile
            
            return (fixed_value, profile)
        
        if isinstance(self.config, dict):
            if customer.cust_type not in self.config:
                print(
                    f"{customer.cust_type} missing from config data {self.config}"
                )
                return (None, None)
            return _get_kw_and_profile(self.config.get(customer.cust_type))
            
        return _get_kw_and_profile(self.config)
    

SIZING_STRATEGY_MAPPER  ={
    CapacityStrategyEnum.solar_energy_based: EnergyBasedSolarSizingStrategy,
    CapacityStrategyEnum.peak_multiplier: PeakMultiplierSizingStrategy,
    CapacityStrategyEnum.fixed_sizing: FixedSizingStrategy
}



"""
Create a geoJSON for a feeder model
"""

# standard imports
from pathlib import Path

# internal imports
from emerge.utils.util import validate_path, write_file

def create_feeder_geojson(
    dss_instance,
    output_folder
):

    output_folder = Path(output_folder)
    output_folder.mkdir(exist_ok=True)
    validate_path(output_folder)

    # Let's get all the buses
    bus_geojson = {"type": "FeatureCollection", "features": []}
    
    # Get all the busnames, loop over them and find the 
    # x,y coordinates 
    all_buses = dss_instance.Circuit.AllBusNames()
    bus_coord_dict = {}
    for bus in all_buses:
        dss_instance.Circuit.SetActiveBus(bus)
        
        x, y = dss_instance.Bus.X(), dss_instance.Bus.Y()
        bus_coord_dict[bus] = {
            'longitude': x,
            'latitude': y
        }

        bus_geojson['features'].append(
            {
                "type": "Feature",
                "geometry": {
                    "type": "Point",
                    "coordinates": [x, y]
                },
                "properties": {
                    "name": bus
                }
            }
        )
    write_file(bus_geojson, output_folder / 'buses.json')

    # Get all the line sections
    lines_geojson = {"type": "FeatureCollection", "features": []}
    flag = dss_instance.Lines.First()
    while flag:
        line_name = dss_instance.Lines.Name()
        bus1, bus2 = dss_instance.Lines.Bus1().split('.')[0], dss_instance.Lines.Bus2().split('.')[0]
        n_customers = dss_instance.Lines.TotalCust()
        lines_geojson['features'].append(
            {
                "type": "Feature",
                "geometry": {
                    "type": "LineString",
                    "coordinates": [
                        [bus_coord_dict[bus1]['longitude'], bus_coord_dict[bus1]['latitude']],
                        [bus_coord_dict[bus2]['longitude'], bus_coord_dict[bus2]['latitude']]
                    ],
                    "properties": {
                        "name": line_name,
                        "n_customers": n_customers,
                        "bus1": bus1,
                        "bus2": bus2
                    }
                }
            }
        )

        flag = dss_instance.Lines.Next()
    
    write_file(lines_geojson, output_folder / 'lines.json')

    # Get all transformers
    transformer_geojson = {"type": "FeatureCollection", "features": []}
    flag = dss_instance.Transformers.First()
    while flag:
        transformer_name = dss_instance.CktElement.Name()
        bus1, bus2 = dss_instance.CktElement.BusNames()[:2]
        bus1, bus2 = bus1.split('.')[0], bus2.split('.')[0]
        transformer_geojson['features'].append(
            {
                "type": "Feature",
                "geometry": {
                    "type": "Point",
                    "coordinates": [bus_coord_dict[bus1]['longitude'], bus_coord_dict[bus1]['latitude']],
                    "properties": {
                        "name": transformer_name,
                        "bus1": bus1,
                        "bus2": bus2,
                        "kva": dss_instance.Transformers.kVA()
                    }
                }
            }
        )

        flag = dss_instance.Transformers.Next()
    
    write_file(transformer_geojson, output_folder / 'transformers.json')

    # Get all loads
    load_geojson = {"type": "FeatureCollection", "features": []}
    flag = dss_instance.Loads.First()
    while flag:
        load_name = dss_instance.CktElement.Name()
        bus1 = dss_instance.CktElement.BusNames()[0]
        bus1 = bus1.split('.')[0]
        load_geojson['features'].append(
            {
                "type": "Feature",
                "geometry": {
                    "type": "Point",
                    "coordinates": [bus_coord_dict[bus1]['longitude'], bus_coord_dict[bus1]['latitude']],
                    "properties": {
                        "name": load_name,
                        "bus": bus1,
                        "kw": dss_instance.Loads.kW(),
                        "kvar": dss_instance.Loads.kvar()
                    }
                }
            }
        )

        flag = dss_instance.Loads.Next()
    
    write_file(load_geojson, output_folder / 'loads.json')




""" This module contains tests for computing system level metrics."""

from emerge.metrics import node_metrics
from emerge.metrics import observer
from conftest import simulation_manager_setup

def test_nvri_metric():
    """ Function to test the computation of NVRI ."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    nvri_observer = node_metrics.NVRI()
    subject.attach(nvri_observer)

    manager.simulate(subject)
    nvri = nvri_observer.get_metric()
    assert nvri

def test_llri_metric():
    """ Function to test the computation of LLRI ."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    llri_observer = node_metrics.LLRI()
    subject.attach(llri_observer)

    manager.simulate(subject)
    llri = llri_observer.get_metric()
    assert llri

def test_tlri_metric():
    """ Function to test the computation of TLRI ."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    tlri_observer = node_metrics.TLRI()
    subject.attach(tlri_observer)

    manager.simulate(subject)
    tlri = tlri_observer.get_metric()
    assert tlri



"""
Higher level wrapper to compute all the metrics
"""

# third-party imports
from pathlib import Path

# internal imports
from emerge.utils.util import setup_logging
from emerge.db.db_handler import TinyDBHandler
from emerge.simulator.opendss import OpenDSSSimulator
from emerge.network.asset_metrics import (aggregate_asset_metrics,
    get_coordinates)
from emerge.simulator.powerflow_results import (
    get_allbus_voltage_pu,
    get_voltage_by_distance,
    get_voltage_by_lat_lon,
    get_voltage_distribution,
    get_lineloading_dataframe,
    get_transloading_dataframe
)

def compute_snapshot_metrics(
    master_dss_file_path: str,
    output_path: str = Path(__file__).parents[2] / 'db.json'
):
    """ Computes metrics for snapshot powerflow results. It takes master dss file and
    computes all the snapshot metrics for the purpose of visualization. This will create a
    db json file which is used by dashboard to plot results """
    
    setup_logging()
    db_instance = TinyDBHandler(output_path)
    opendss_instance = OpenDSSSimulator(master_dss_file_path)
    opendss_instance.solve()
    
    """ Get all the asset metrics """
    asset_metrics = aggregate_asset_metrics(opendss_instance.dss_instance).dict()
    db_instance.db.insert({"type": "asset_metrics", "metrics": asset_metrics})
    
    """ Get coordinates """
    coordinates = get_coordinates(opendss_instance.dss_instance)
    db_instance.db.insert({"type": "coordinates", "data": coordinates})

    """ Get voltage pu distributions """
    voltages = get_allbus_voltage_pu(opendss_instance.dss_instance)
    voltage_bins = get_voltage_distribution(voltages)
    db_instance.db.insert({"type": "snapshot_voltage_bins", \
        "label": "peak_load", "data": voltage_bins })

    """ Get voltage pu by disyance"""
    voltage_by_distance = get_voltage_by_distance(opendss_instance.dss_instance)
    db_instance.db.insert({"type": "snapshot_voltage_by_distance", \
        "label": "peak_load", "data":  voltage_by_distance})

    """ Get voltage by lat lons"""
    voltage_by_lat_lon = get_voltage_by_lat_lon(opendss_instance.dss_instance)
    db_instance.db.insert({
            "type": "snapshot_voltage_for_heatmap", 
            "label": "peak_load", 
            "data":  voltage_by_lat_lon})

    """ line loading """
    line_loading_df = get_lineloading_dataframe(opendss_instance.dss_instance)
    db_instance.db.insert({
            "type": "snapshot_lineloading_for_heatmap", 
            "label": "peak_load", 
            "data":  line_loading_df.to_dict()['loading(pu)']})

    """ transformer loading """
    xfmr_loading_df = get_transloading_dataframe(opendss_instance.dss_instance)
    db_instance.db.insert({
            "type": "snapshot_xfmrloading_for_heatmap", 
            "label": "peak_load", 
            "data":  xfmr_loading_df.to_dict()['loading(pu)']})




""" Module for handling testing of utility functions. """

from pathlib import Path

import pandas as pd

from emerge.utils import dss_util
from emerge.simulator import opendss
from emerge.api import utils


def test_bus_distance_dataframe():
    """Test function for checking computation
    of bus distance dataframe."""

    root_path = Path(__file__).absolute().parents[1]
    master_dss_file = root_path / "examples" / "opendss" / "master.dss"

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    bus_distance_df = dss_util.get_bus_distance_dataframe(
        simulator.dss_instance
    )

    assert isinstance(bus_distance_df, pd.DataFrame)


def test_get_list_of_customer_models():
    """Test function for getting list of customer models."""

    root_path = Path(__file__).absolute().parents[1]
    master_dss_file = root_path / "examples" / "opendss" / "master.dss"

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    customer_models = dss_util.get_list_of_customer_models(
        simulator.dss_instance
    )



def test_get_load_mapper_objects():
    """Test function for getting load mapper objects."""

    root_path = Path(__file__).absolute().parents[1]
    master_dss_file = root_path / "examples" / "opendss" / "master.dss"

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    mapper_objects = dss_util.get_load_mapper_objects(simulator.dss_instance)



def test_get_line_customers():
    """ Test function for `get_line_customers`
    utility function."""

    root_path = Path(__file__).absolute().parents[1]
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    line_customers_df = dss_util.get_line_customers(
        simulator.dss_instance
    )

    assert isinstance(line_customers_df, pd.DataFrame)

def test_get_transformer_customers():
    """ Test function for `get_transformer_customers`
    utility function."""

    root_path = Path(__file__).absolute().parents[1]
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    xfmr_customers_df = dss_util.get_transformer_customers(
        simulator.dss_instance
    )

    assert isinstance(xfmr_customers_df, pd.DataFrame)

def test_get_source_node():
    """ test function for getting source node. """
    root_path = Path(__file__).absolute().parents[1]
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    source_node = dss_util.get_source_node(
        simulator.dss_instance
    )
    assert source_node == '80_27834864364979_13_0915193376659_htnode'

def test_get_bus_load_flag():
    """ Test function for getting mapping between bus and load flag. """

    root_path = Path(__file__).absolute().parents[1]
    master_dss_file = root_path / 'examples' / 'opendss' / 'master.dss'

    simulator = opendss.OpenDSSSimulator(master_dss_file)
    is_bus_load_df = dss_util.get_bus_load_flag(
        simulator.dss_instance
    )
    assert is_bus_load_df.sum()['is_load'] == simulator.dss_instance.Loads.Count()


def test_buses_coordinate_mapping():
    """ Test function for getting  buses coordinate mapping. """

    base_path = Path(__file__).absolute().parents[0] / 'data'
    bus_to_coordinate_mapping = utils.buses_coordinate_mapping(base_path / 'buses.json')
    assert isinstance(bus_to_coordinate_mapping, dict)

def test_lines_coordinate_mapping():
    """ Test function for getting lines coordinate mapping. """

    base_path = Path(__file__).absolute().parents[0] / 'data'
    lines_to_coordinate_mapping = utils.lines_coordinate_mapping(base_path / 'lines.json')
    assert isinstance(lines_to_coordinate_mapping, dict)

def test_transformer_coordinate_mapping():
    """ Test function for getting transformer coordinate mapping. """

    base_path = Path(__file__).absolute().parents[0] / 'data'
    xfmr_to_coordinate_mapping = utils.transformer_coordinate_mapping(base_path / 'transformers.json')
    assert isinstance(xfmr_to_coordinate_mapping, dict)

def test_getting_map_center():
    """ Test function for getting map center. """

    base_path = Path(__file__).absolute().parents[0] / 'data'
    map_center = utils.get_map_center(base_path / 'buses.json')
    assert isinstance(map_center, dict)



""" Module for managing EMERGE specific exceptions. """

class BaseEmergeException(Exception):
    """ Base class for managing emerge exceptions. """

class EnergyMeterNotDefined(Exception):
    """ Exception raised because energy meter is not defined. """


""" Model for input config for API. """

from pydantic import BaseModel


class Config(BaseModel):
    snapshot_metrics_db: str
    timeseries_metrics_db: str
    geojson_path: str
    scenario_metrics_db: str
    scenario_metrics_vvar: str
    scenario_metrics_vvar_nighttime: str
    ui_url: str



""" Module for testing opendss time series simulation. """

from pathlib import Path

from conftest import simulation_manager_setup

def test_timeseries_simulation():
    """ Function for tesitng basic time series simulation capability"""

    manager = simulation_manager_setup()
    manager.simulate()
    manager.export_convergence()

    convergence_file = Path('convergence.csv')
    assert convergence_file.exists()
    convergence_file.unlink()


""" Module for managing metric computation subscriber and publisher. """

import abc
import uuid
from typing import Dict, List
import polars

from emerge.db import db_handler

class MetricObserver(abc.ABC):
    """ Abstracte interace for metrics observers"""

    _id = str(uuid.uuid4())

    @abc.abstractmethod
    def compute(self, *args, **kwargs)-> None:
        """ All metric observer subclass must implement compute method. """

    @abc.abstractmethod
    def get_metric(self)-> Dict:
        """ All metric observer subclass must implement get_metric method. """


class MetricsSubject:
    """ Class for managing metric subscribers """

    _subscribers = []

    def _observer_exists(self, observer: MetricObserver):
        """ Check whether the observer exists or not"""
        
        for id, obs in enumerate(self._subscribers):
            if obs._id == observer._id:
                return id

        return False

    def attach(self, observer: MetricObserver):
        """ Method for attaching the observers. """
        if not self._observer_exists(observer):
            self._subscribers.append(observer)

    def detach(self, observer: MetricObserver):
        """ Method for deleting the observer object from the list. """
        
        observer_index = self._observer_exists(observer)
        if observer_index:
            self._subscribers.pop(observer_index)


    def notify(self, *args, **kwargs):
        """ Method for notifying the observers. """
        for obs in self._subscribers:
            obs.compute(*args, **kwargs)

    
def export_tinydb_json(observers: List[MetricObserver], json_path: str ):
    """ Function for exporting metrics. """
    
    db_instance = db_handler.TinyDBHandler(json_path)
    for observer in observers:
        db_instance.db.insert(
            {
                "type": "metrics", 
                "name": observer.__class__.__name__,
                "data": observer.get_metric()
            })

def export_csv(observers: List[MetricObserver], export_path):
    
    for observer in observers:
        df = polars.from_dict(observer.get_metric())
        df.write_csv(export_path / (str(observer.__class__.__name__) + '.csv'))


""" This module contains tests for computing system level metrics."""

from emerge.metrics import node_voltage_stats
from emerge.metrics import observer
from conftest import simulation_manager_setup

def test_node_voltage_stats_metric():
    """ Function to test the computation of Node voltage statistical metrics."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    node_v_obs = node_voltage_stats.NodeVoltageStats()
    subject.attach(node_v_obs)

    manager.simulate(subject)
    node_v = node_v_obs.get_metric()
    print(node_v)


def test_node_voltage_bins():
    """ Function to test the computation of Node voltage bins."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    node_v_obs = node_voltage_stats.NodeVoltageBins(
        [0.95, 0.98, 1.02, 1.05]
    )
    subject.attach(node_v_obs)

    manager.simulate(subject)
    node_v = node_v_obs.get_metric()
    print(node_v)


# standard imports
from pathlib import Path

# third-party imports
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from tinydb import Query
import numpy as np

# internal imports
from emerge.utils.util import read_file
from emerge.db.db_handler import TinyDBHandler
from emerge.api import config_model
from emerge.api import utils

config_json = Path(__file__).absolute().parents[0] / "config.json"
config = config_model.Config.parse_file(config_json)

buses_to_coordinate_mapping = utils.buses_coordinate_mapping(
    Path(config.geojson_path) / "buses.json"
)
lines_to_coordinate_mapping = utils.lines_coordinate_mapping(
    Path(config.geojson_path) / "lines.json"
)
xfmr_to_coordinate_mapping = utils.lines_coordinate_mapping(
    Path(config.geojson_path) / "transformers.json"
)
map_center = utils.get_map_center(Path(config.geojson_path) / "buses.json")
query = Query()
db_snapshot = TinyDBHandler(config.snapshot_metrics_db)
db_timeseries = TinyDBHandler(config.timeseries_metrics_db)

scenario_metrics = {
    "unity_pf": {},
    "voltvar": {},
    "voltvar_night_off": {}
}
scenario_paths = {
    "unity_pf": Path(config.scenario_metrics_db),
    "voltvar": Path(config.scenario_metrics_vvar),
    "voltvar_night_off": Path(config.scenario_metrics_vvar_nighttime)
}
for scenario_name, scenario_path in scenario_paths.items():
    for filepath in scenario_path.iterdir():
        if filepath.suffix == ".json":
            scenario_metrics[scenario_name][filepath.name.split(".")[0]] = TinyDBHandler(
                filepath
            )
        
    scenario_metrics[scenario_name]['scenario_0_0'] = TinyDBHandler(config.timeseries_metrics_db)


# Fast API setup
app = FastAPI()
origins = [config.ui_url]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/scenarios/timeseries_asset/{metric_name}")
def get_scenario_system_metrics(metric_name: str):

    json_content = []
    
    for func in [{"name": "max", "func": np.nanmax}, 
                {"name": "min", "func": np.nanmin},
                {"name": "90-percentile", "func": np.percentile, "args": [90]}, 
                {"name": "mean", "func": np.nanmean},
                {"name": "98-percentile", "func": np.percentile, "args": [98]},
                {"name": "99-percentile", "func": np.percentile, "args": [99]}]:
        
        metrics = []
        for name, db_ in scenario_metrics["unity_pf"].items():
            
            query_result = db_.db.search(
                query.type == "metrics" and query.name == metric_name
            )
            
            if query_result:
                query_result = query_result[0]["data"]
                
                metric_values = list(query_result.values())
                metrics.append(
                    {"metric": func["func"](metric_values) if "args"  not in \
                        func else func["func"](metric_values, *func['args']), 
                    "name": name.split("_")[2] + "%"}
                )
            
        if metrics:
            metrics = sorted(
                metrics, key=lambda d: float(d["name"].split("%")[0])
            )
        

            json_content.append(
                {
                    "type": "scatter",
                    "x": [metric["name"] for metric in metrics],
                    "y": [metric["metric"] for metric in metrics],
                    "name": func["name"]
                },
            )
            

    return json_content


@app.get("/scenarios/timeseries/{metric_name}")
def get_scenario_system_metrics(metric_name: str):

    json_content = []
    for which_power in ['active_power', 'reactive_power']:
        for scenario_name, scenario_metrics_db in scenario_metrics.items():
            metrics = []
            
            for name, db_ in scenario_metrics_db.items():
                metric = db_.db.search(
                    query.type == "metrics" and query.name == metric_name
                )

                if metric:
                    metric = metric[0]["data"][which_power]
                    metrics.append(
                        {"metric": metric, "name": name.split("_")[2] + "%"}
                    )
            

            if metrics:
                metrics = sorted(
                    metrics, key=lambda d: float(d["name"].split("%")[0])
                )
        

                json_content.append(
                    {
                        "type": "scatter",
                        "x": [metric["name"] for metric in metrics],
                        "y": [metric["metric"] for metric in metrics],
                        "name": scenario_name + '__' + which_power
                    },
                )
                

    return json_content


@app.get("/scenarios/system_metrics")
def get_scenario_system_metrics():
    
    json_content = []
    
    for metric_name in ['SARDI_voltage', 'SARDI_aggregated', 'SARDI_line', 'SARDI_transformer']:
        
        for scenario_name, scenario_metrics_db in scenario_metrics.items():
            metrics = []
            for name, db_ in scenario_metrics_db.items():
                
                metric = db_.db.search(
                    query.type == "metrics" and query.name == metric_name
                )

                if metric:
                    metric = metric[0]["data"][metric_name.lower()]

                    metrics.append(
                        {"metric": metric, "name": name.split("_")[2] + "%"}
                    )
            
            if metrics:
                metrics = sorted(
                    metrics, key=lambda d: float(d["name"].split("%")[0])
                )
        

                json_content.append(
                    {
                        "type": "scatter",
                        "x": [metric["name"] for metric in metrics],
                        "y": [metric["metric"] for metric in metrics],
                        "name": scenario_name + '__' + metric_name 
                    },
                )
                

    return json_content


@app.get("/assets/geojsons/buses")
def get_buses_geojson():

    file_path = Path(config.geojson_path) / "buses.json"
    json_content = {
        "type": file_path.name.split(".")[0],
        "data": read_file(file_path),
    }

    return json_content


@app.get("/assets/geojsons/lines")
def get_lines_geojson():

    file_path = Path(config.geojson_path) / "lines.json"
    json_content = {
        "type": file_path.name.split(".")[0],
        "data": read_file(file_path),
    }

    return json_content


@app.get("/assets/geojsons/transformers")
def get_transformer_geojson():

    file_path = Path(config.geojson_path) / "transformers.json"
    json_content = {
        "type": file_path.name.split(".")[0],
        "data": read_file(file_path),
    }

    return json_content


@app.get("/assets/geojsons/loads")
def get_loads_geojson():

    file_path = Path(config.geojson_path) / "loads.json"
    json_content = {
        "type": file_path.name.split(".")[0],
        "data": read_file(file_path),
    }

    return json_content


@app.get("/assets/metrics")
def get_asset_metrics():
  
    asset_metrics = db_snapshot.db.search(query.type == "asset_metrics")[0][
        "metrics"
    ]

    json_content = [
        {
            "type": key,
            "data": [
                {"metric": m.replace("_", " "), "value": round(v, 2)}
                for m, v in value.items()
            ],
        }
        for key, value in asset_metrics.items()
    ]

    return json_content

@app.get("/snapshots/line_loading")
def get_snapshot_line_loading():

    metric_data = db_snapshot.db.search(
        query.type == "snapshot_lineloading_for_heatmap"
    )[0]["data"]

    json_content = [
        {
            "coordinates": lines_to_coordinate_mapping[key.split(".")[1]],
            "name": key,
            "data": value,
        }
        for key, value in metric_data.items()
    ]

    return json_content

@app.get("/snapshots/xfmr_loading")
def get_snapshot_xfmr_loading():
 
    metric_data = db_snapshot.db.search(
        query.type == "snapshot_xfmrloading_for_heatmap"
    )[0]["data"]

    json_content = [
        {
            "coordinates": xfmr_to_coordinate_mapping[key.replace('transformer.', 'Transformer.')],
            "weight": value
        }
        for key, value in metric_data.items()
    ]

    return json_content

@app.get("/snapshots/voltage")
def get_snapshots_voltage():
   
    voltage_for_heatmap = db_snapshot.db.search(
        query.type == "snapshot_voltage_for_heatmap"
    )[0]["data"]
    json_content = [
        {"coordinates": [lon, lat], "weight": voltage}
        for lat, lon, voltage in zip(
            voltage_for_heatmap["latitudes"],
            voltage_for_heatmap["longitudes"],
            voltage_for_heatmap["voltage (pu)"],
        )
    ]

    return json_content


@app.get("/snapshots/voltage-distribution")
def get_snapshots_voltage():
  
    voltage_bins = db_snapshot.db.search(query.type == "snapshot_voltage_bins")
    json_content = [
        {
            "x": arr["data"]["Voltage bins"],
            "y": arr["data"]["Percentage nodes"],
            "name": arr["label"],
            "type": "bar",
        }
        for arr in voltage_bins
    ]

    return json_content

@app.get("/snapshots/voltage-by-distance")
def get_voltage_by_distance():

    voltage_by_distance = db_snapshot.db.search(query.type == "snapshot_voltage_by_distance")[0]

    voltage_by_phase = {}
    json_content = []
    for distance, phase, pu in zip(voltage_by_distance["data"]["Distance from substation (km)"],
        voltage_by_distance["data"]["Phase"], voltage_by_distance["data"]["Voltage (pu)"]):
        if phase not in voltage_by_phase:
            voltage_by_phase[phase] = {
                "distance": [],
                "pu": []
            }
        voltage_by_phase[phase]['distance'].append(distance)
        voltage_by_phase[phase]['pu'].append(pu)

    for key, subdict in voltage_by_phase.items():
        json_content.append(
            {
                "x": subdict['distance'],
                "y": subdict['pu'],
                "name": f"phase_{key}",
                "mode": 'markers',
                "type": "scatter"
            }
        )

    return json_content



@app.get("/metrics/timeseries/nvri")
def get_timeseries_metric():
 
    metric_data = db_timeseries.db.search(
        query.type == "metrics" and query.name == "NVRI"
    )[0]["data"]

    json_content = [
        {
            "coordinates": buses_to_coordinate_mapping[key],
            "name": key,
            "data": value,
        }
        for key, value in metric_data.items()
    ]

    return json_content




@app.get("/metrics/timeseries/llri")
def get_timeseries_metric():
    
    metric_data = db_timeseries.db.search(
        query.type == "metrics" and query.name == "LLRI"
    )[0]["data"]

    json_content = [
        {
            "coordinates": lines_to_coordinate_mapping[key.split(".")[1]],
            "name": key,
            "data": value,
        }
        for key, value in metric_data.items()
    ]

    return json_content


@app.get("/metrics/timeseries/tlri")
def get_timeseries_metric():
  
    metric_data = db_timeseries.db.search(
        query.type == "metrics" and query.name == "TLRI"
    )[0]["data"]

    json_content = [
        {
            "coordinates": xfmr_to_coordinate_mapping[
                "Transformer." + key.split(".")[1]
            ],
            "name": key,
            "data": value,
        }
        for key, value in metric_data.items()
    ]

    return json_content


@app.get("/metrics/system_metrics")
def get_system_metric():
    
    metrics = [
        "SARDI_voltage",
        "SARDI_line",
        "SARDI_transformer",
        "SARDI_aggregated",
    ]

    data = []

    for metric in metrics:
        metric_data = db_timeseries.db.search(
            query.type == "metrics" and query.name == metric
        )[0]["data"]
        data.append(
            {
                "name": metric.replace("_", " "),
                "value": metric_data[metric.lower()],
            }
        )

    return data


@app.get("/map_center")
def get_map_center():
    return map_center





""" This module contains tests for computing system level metrics."""
from pathlib import Path

from emerge.metrics import system_metrics
from emerge.metrics import observer
from conftest import simulation_manager_setup

def test_total_energy():
    """ Function to test the computation of total energy."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    total_energy_observer = system_metrics.TotalEnergy()
    subject.attach(total_energy_observer)

    manager.simulate(subject)

def test_pv_energy():
    """ Function to test the computation of total PV energy."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    pv_observer = system_metrics.TotalPVGeneration()
    subject.attach(pv_observer)

    manager.simulate(subject)

def test_energy_loss():
    """ Function to test the computation of total loss."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    loss_observer = system_metrics.TotalLossEnergy()
    subject.attach(loss_observer)

    manager.simulate(subject)
    print(loss_observer.get_metric())

def test_timeseries_powerloss():
    """ Function to test the computation of timeseries power loss."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    loss_observer = system_metrics.TimeseriesTotalLoss()
    subject.attach(loss_observer)

    manager.simulate(subject)
    print(loss_observer.get_metric())

def test_sardi_voltage():
    """ Function to test the computation of SARDI voltage."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    sardi_voltage_observer = system_metrics.SARDI_voltage()
    subject.attach(sardi_voltage_observer)

    manager.simulate(subject)

def test_sardi_line():
    """ Function to test the computation of SARDI line."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    sardi_line_observer = system_metrics.SARDI_line()
    subject.attach(sardi_line_observer)

    manager.simulate(subject)

def test_sardi_transformer():
    """ Function to test the computation of SARDI transformer."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    sardi_xfmr_observer = system_metrics.SARDI_transformer()
    subject.attach(sardi_xfmr_observer)

    manager.simulate(subject)
    print(sardi_xfmr_observer.get_metric())

def test_sardi_aggregated():
    """ Function to test the computation of SARDI aggregated."""

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    sardi_aggregated_observer = system_metrics.SARDI_aggregated()
    subject.attach(sardi_aggregated_observer)

    manager.simulate(subject)


def test_sardi_metrics():
    """ Function for testing SARDI metrics. """

    manager = simulation_manager_setup()
    pvsystem_file = Path(__file__).absolute().parents[1]/ 'examples' / 'opendss' / 'PVSystems.dss'
    manager.opendss_instance.execute_dss_command(f"redirect {pvsystem_file}")
    subject = observer.MetricsSubject()
    sardi_voltage_observer = system_metrics.SARDI_voltage()
    sardi_line_observer = system_metrics.SARDI_line()
    sardi_aggregated_observer = system_metrics.SARDI_aggregated()
    sardi_xfmr_observer = system_metrics.SARDI_transformer()
    
    subject.attach(sardi_voltage_observer)
    subject.attach(sardi_line_observer)
    subject.attach(sardi_xfmr_observer)
    subject.attach(sardi_aggregated_observer)

    manager.simulate(subject)

    sardi_voltage = list(sardi_voltage_observer.get_metric().values())[0]
    sardi_line = list(sardi_line_observer.get_metric().values())[0]
    sardi_xfmr = list(sardi_xfmr_observer.get_metric().values())[0]
    sardi_aggregated = list(sardi_aggregated_observer.get_metric().values())[0]

    assert sardi_aggregated >= sardi_voltage
    assert sardi_aggregated >= sardi_line
    assert sardi_aggregated >= sardi_xfmr



""" Utility functions for dss_instance. """

import copy
from typing import List

import opendssdirect as dss
import pandas as pd
import networkx as nx

from emerge.metrics.exceptions import EnergyMeterNotDefined
from emerge.network import asset_metrics
from emerge.scenarios import data_model


def get_bus_distance_dataframe(dss_instance: dss)-> pd.DataFrame:
    """Get bus distance dataframe.
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        pd.DataFrame: Dataframe containing distance from 
            substation indexed by busname.
    
    """

    bus_distance_df = {"busname": [], "distance": []}
    for bus in dss_instance.Circuit.AllBusNames():
        dss_instance.Circuit.SetActiveBus(bus)

        bus_distance_df["busname"].append(bus)
        bus_distance_df["distance"].append(dss_instance.Bus.Distance())

    if not sum(bus_distance_df["distance"]):
        raise EnergyMeterNotDefined(
            "Define energy meter to be able to find"
            "number of customers downward of line segment."
        )
    return pd.DataFrame(bus_distance_df).set_index("busname")


def get_list_of_customer_models(
        dss_instance: dss, 
        load_multiplier=1.0,
        cust_type: str = "yearly"
    ) -> List[data_model.CustomerModel]:
    """Returns list of customer models from dss instance.
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        List[data_model.CustomerModel]: List of customer data
            model.
    """
    cust_type_func = {
        "class": dss_instance.Loads.Class,
        "yearly": dss_instance.Loads.Yearly
    }.get(cust_type)

    flag = dss_instance.Loads.First()
    bus_distance_df = get_bus_distance_dataframe(dss_instance)
    customer_model = []
    while flag:
        busname = dss_instance.CktElement.BusNames()[0].split('.')[0]
        customer_model.append(
            data_model.CustomerModel(
                name=dss_instance.CktElement.Name(),
                kw=dss_instance.Loads.kW()*load_multiplier,
                distance=bus_distance_df.loc[busname]['distance'],
                cust_type=str(cust_type_func())
            )
        )

        flag = dss_instance.Loads.Next()

    return customer_model


def get_load_mapper_objects(dss_instance: dss)-> List[data_model.LoadMetadataModel]:
    """Returns list of load mapper object models.
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        List[data_model.LoadMetadataModel]: List of load
            metadata model.
    """

    flag = dss_instance.Loads.First()
    mapper_model = []
    while flag:
        mapper_model.append(
            data_model.LoadMetadataModel(
                name=dss_instance.CktElement.Name(),
                bus=dss_instance.CktElement.BusNames()[0],
                num_phase=dss_instance.Loads.Phases(),
                kv=dss_instance.Loads.kV(),
                yearly=dss_instance.Loads.Yearly()
            )
        )

        flag = dss_instance.Loads.Next()

    return mapper_model


def get_bus_load_dataframe(dss_instance: dss)-> pd.DataFrame:
    """Bus to load mapping dataframe.
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        pd.DataFrame: Dataframe containing mapping between load
            name and bus name
    """

    flag = dss_instance.Loads.First()
    load_bus_name_dict = {"busname": [], "loadname": []}
    while flag:

        load_name = dss_instance.CktElement.Name()
        busname = dss_instance.CktElement.BusNames()[0].split(".")[0]
        load_bus_name_dict["busname"].append(busname)
        load_bus_name_dict["loadname"].append(load_name)

        flag = dss_instance.Loads.Next()

    return pd.DataFrame(load_bus_name_dict)


def get_bus_load_flag(dss_instance: dss) -> pd.DataFrame:
    """Bus to load mapping dataframe.

    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        pd.DataFrame: Dataframe containing load flag
            indexed by bus name.
    """

    load_bus_df = get_bus_load_dataframe(dss_instance)
    buses_with_load = list(load_bus_df["busname"])
    all_buses = dss_instance.Circuit.AllBusNames()
    buses_without_load = list(set(all_buses) - set(buses_with_load))
    is_bus_load = {
        "busname": buses_with_load + buses_without_load,
        "is_load": [1] * len(buses_with_load) + [0] * len(buses_without_load),
    }

    return pd.DataFrame(is_bus_load).set_index("busname")


def get_line_customers(dss_instance: dss) ->pd.DataFrame:
    """Function to retrieve dataframe containing number
    of downward serving customers for all line segments.

    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        pd.DataFrame: Dataframe containing number of customers
            indexed by line name.
    """
    line_customers_df = {"linename": [], "customers": []}

    dss_instance.Circuit.SetActiveClass("Line")
    flag = dss_instance.ActiveClass.First()

    while flag:
        line_name = dss_instance.CktElement.Name().lower()
        n_customers = dss_instance.Lines.TotalCust()
        line_customers_df["linename"].append(line_name)
        line_customers_df["customers"].append(n_customers)
        flag = dss_instance.ActiveClass.Next()

    if not any(line_customers_df["customers"]):
        raise EnergyMeterNotDefined(
            "Define energy meter to be able to find"
            "number of customers downward of line segment."
        )
    return pd.DataFrame(line_customers_df).set_index("linename")


def get_transformer_customers(dss_instance: dss)-> pd.DataFrame:
    """Function to retrieve dataframe containing number
    of downward serving customers for all transformers.

    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        pd.DataFrame: Dataframe containing number of customers
            indexed by transformer name.
    """
    xfmr_customers_df = {"transformername": [], "customers": []}

    network = asset_metrics.networkx_from_opendss_model(dss_instance)
    substation_bus = get_source_node(dss_instance)
    bus_load_flag_df = get_bus_load_flag(dss_instance)

    dss_instance.Circuit.SetActiveClass("Transformer")
    flag = dss_instance.ActiveClass.First()

    while flag:
        network_copy = copy.deepcopy(network)
        xmfr_name = dss_instance.CktElement.Name().lower()

        edge_to_be_removed = []
        for edge in network_copy.edges():
            edge_data = network_copy.get_edge_data(*edge)
            if (
                edge_data
                and "name" in edge_data
                and edge_data["name"] == xmfr_name
            ):
                edge_to_be_removed.append(edge)

        network_copy.remove_edges_from(edge_to_be_removed)
        connected_buses = nx.node_connected_component(
            network_copy, substation_bus
        )

        filtered_bus_df = bus_load_flag_df.loc[
            bus_load_flag_df.index.difference(connected_buses)
        ]

        n_customers = filtered_bus_df.sum()["is_load"]
        xfmr_customers_df["transformername"].append(xmfr_name)
        xfmr_customers_df["customers"].append(n_customers)
        flag = dss_instance.ActiveClass.Next()

    return pd.DataFrame(xfmr_customers_df).set_index("transformername")


def get_source_node(dss_instance: dss) -> str:

    """Function to return source node for dss model.
    
    Args:
        dss_instance (dss): Instance of OpenDSSDirect

    Returns:
        str: Name of the source node
    """

    df = dss_instance.utils.class_to_dataframe("vsource")
    source_node = df["bus1"].tolist()[0].split(".")[0]
    return source_node



""" Modules containing test for exporting metrics. """
import os

from emerge.metrics import system_metrics
from emerge.metrics import observer
from emerge.metrics import node_metrics
from conftest import simulation_manager_setup

def test_export_metrics():
    """ Test exporting metrics. """

    manager = simulation_manager_setup()
    subject = observer.MetricsSubject()

    sardi_voltage_observer = system_metrics.SARDI_voltage()
    sardi_line_observer = system_metrics.SARDI_line()
    sardi_xfmr_observer = system_metrics.SARDI_transformer()
    sardi_aggregated_observer = system_metrics.SARDI_aggregated()
    nvri_observer = node_metrics.NVRI()
    llri_observer = node_metrics.LLRI()
    tlri_observer = node_metrics.TLRI()
    
    observers_ = [sardi_voltage_observer, sardi_line_observer,
        sardi_xfmr_observer, sardi_aggregated_observer, nvri_observer,
        llri_observer, tlri_observer]
    for observer_ in observers_:
        subject.attach(observer_)

    manager.simulate(subject)

    observer.export_tinydb_json(observers_, "db_metrics.json")
    os.remove("db_metrics.json")



""" Module containing utility functions for API. """

from emerge.utils.util import read_file

def get_map_center(buses_file_path: str):
    """ Function to get map centre from bus geojson file. """

    buses_geojson = read_file(buses_file_path)
    longitudes, latitudes = [], []
    for feature in buses_geojson['features']:
        coordinates = feature["geometry"]["coordinates"]
        longitudes.append(coordinates[0])
        latitudes.append(coordinates[1])

    return {
        'longitude': sum(longitudes)/len(longitudes) if longitudes else 0,
        'latitude': sum(latitudes)/len(latitudes) if latitudes else 0
    }

def buses_coordinate_mapping(buses_file_path: str):
    """ Function to get mapping formbus and coordinate. """

    buses_geojson = read_file(buses_file_path)
    buses_mapping  = {}
    for feature in buses_geojson['features']:
        bus_name = feature["properties"]["name"]
        coordinates = feature["geometry"]["coordinates"]
        buses_mapping[bus_name] = coordinates

    return buses_mapping

def lines_coordinate_mapping(lines_geojson_path: str):
    """ Function to compute mapping for line segment and coordinate. """
    
    lines_geojson = read_file(lines_geojson_path)
    lines_mapping  = {}
    for feature in lines_geojson['features']:
        line_name = feature["geometry"]["properties"]["name"]
        coordinates = feature["geometry"]["coordinates"]
        lines_mapping[line_name] = coordinates

    return lines_mapping

def transformer_coordinate_mapping(transformer_geojson_path: str):
    """ Function to compute mapping for transformer and coordinate. """
    
    xfmr_geojson = read_file(transformer_geojson_path)
    xfmr_mapping  = {}
    for feature in xfmr_geojson['features']:
        trans_name = feature["geometry"]["properties"]["name"]
        coordinates = feature["geometry"]["coordinates"]
        xfmr_mapping[trans_name] = coordinates

    return xfmr_mapping




""" Basic utility functions. """

# standard imports
from typing import List, Dict, Union
from pathlib import Path
import logging
import logging.config
import json
import json5

# third-party imports
import yaml

logger = logging.getLogger(__name__)

def get_map_centre(longitudes: List[float], latitudes: List[float])-> Dict:
    """ Returns map centre by taking longitudes and latitudes. """

    return {'lon': sum(longitudes)/len(longitudes), \
        'lat': sum(latitudes)/len(latitudes)}

def validate_path(
        path: str, 
        check_for_dir: bool =True, 
        check_for_file: bool =False, 
        file_types: Union[None, List[str]]=None
    )-> None:
    """ Check for path validation.
    
    Args:
        path (str): Path for either file or folder
        check_for_dir (bool): Set to true if need to check for directory
        check_for_file (bool): Set to true if need to check of
            existence of file
        file_types (Union[None, List[str]]): List of file types against
            which file path is to be checked.

    Raises:
        Exception: If file path does not exist or folder path does 
            not exist or passed file type do not match what is in 
            file_types list.
    """

    if not file_types:
        file_types = []

    path = Path(path)

    if not path.exists():
        raise Exception(f"{path} does not exist!")

    if check_for_dir and not path.is_dir():
        raise Exception(f"{path} is not a directory!")

    if check_for_file and path.suffix not in file_types:
        raise Exception(f"Invalid file type passed {path} , \
            valid file types are {file_types}")

def read_file(file_path: str, use_json5=False) -> Dict:
    """ Reads a file and returns a python dictionary. """

    file_path = Path(file_path)
    logger.debug(f"Attempting to read {file_path}")

    validate_path(file_path, check_for_dir=False, check_for_file=True, \
        file_types=['.json', '.yaml'])

    # Handle JSON file read
    if file_path.suffix == '.json':
        with open(file_path, "r") as f:
            content = json.load(f) if not use_json5 else json5.load(f)

    # Handle YAML file read
    elif file_path.suffix == '.yaml':
        with open(file_path, "r") as f:
            content = yaml.safe_load(f)

    else:
        logger.error(f"""Could not read the {file_path}, this feature is not yet 
        implemented""")
        raise Exception(f"File of type {file_path.suffix} \
            is not yet implemented for reading purpose")

    logger.debug(f"{file_path} read successfully")
    return content

def write_file(content:dict, file_path: str, use_json5=False, **kwargs) -> None:
    """ Takes a python dictionary and writes it into a file. """
    file_path = Path(file_path)
    validate_path(file_path.parent)

    # Handle JSON file write
    if file_path.suffix == '.json':
        with open(file_path, "w") as f:
            json.dump(content, f, **kwargs) if not use_json5 else json5.dump(content, f, **kwargs)

    # Handle YAML file write
    elif file_path.suffix == '.yaml':
        with open(file_path, "w") as f:
            yaml.safe_dump(content, f, **kwargs)

    else:
        raise Exception(f"File of type {file_path.suffix} \
            is not yet implemented for writing purpose")

def setup_logging(filename: str = None):
    """Creates log directory and sets up logging via logging.yaml. """
    
    if filename is None:
        filename = Path(__file__).parents[2] / 'logging.yaml' 

    logging.config.dictConfig(read_file(filename))


""" Module for managing pydantic data models."""
# pylint: disable=no-name-in-module
from pydantic import Field, BaseModel
from typing_extensions import Annotated


class ThermalLoadingLimit(BaseModel):
    """ Model representing thermal loading limit. 
    Example:

        >>> from emerge.metrics.data_model import ThermalLoadingLimit
        >>> ThermalLoadingLimit(threshold=1.0)
    """

    threshold: Annotated[float, Field(ge = 0.0, le=2.0)] = 1.0

class VoltageViolationLimit(BaseModel):
    """ Model representing voltage violation limit. """

    overvoltage_threshold: Annotated[float, Field(ge = 1.01, le=2.0)] = 1.05
    undervoltage_threshold: Annotated[float, Field(ge = 0, le=1.0)] = 0.95


""" This module manages time series simulation for computing 
time series metrics. """

from typing import Union
import datetime
import logging

import pandas as pd

from emerge.simulator import opendss
from emerge.metrics import observer


logger = logging.getLogger(__name__)
logging.basicConfig(format="%(asctime)s %(levelname)s %(message)s", level=logging.DEBUG)

class OpenDSSSimulationManager:
    """ Class for managing time series simulation using OpenDSS.
    
    Attributes:
        path_to_master_dss_file (str): Path to Master.dss file
        simulation_start_time (datetime.datetime): Datetime indicating
            simulation start time
        profile_start_time (datetime.datetime): Datetime indicating
            profile start time
        simulation_end_time (datetime.datetime): Datetime indicating
            when the simulation finishes
        simulation_timestep_min (float): Simulation time resolution
            in minutes
    """

    def __init__(
        self,
        path_to_master_dss_file: str,
        simulation_start_time: datetime.datetime,
        profile_start_time: datetime.datetime,
        simulation_end_time: datetime.datetime,
        simulation_timestep_min: float,
    )-> None:
        
        self.path_to_master_dss_file = path_to_master_dss_file
        self.simulation_start_time = simulation_start_time
        self.profile_start_time = profile_start_time
        self.simulation_end_time = simulation_end_time
        self.simulation_timestep_min = simulation_timestep_min

        self.opendss_instance = opendss.OpenDSSSimulator(self.path_to_master_dss_file)
        self.opendss_instance.set_mode(2)
        self.opendss_instance.set_simulation_time(
            self.simulation_start_time,
            self.profile_start_time
        )
        self.opendss_instance.set_stepsize(self.simulation_timestep_min)
        self.opendss_instance.set_max_iteration(1000)
        self.current_time = self.simulation_start_time
        self.convergence_dict  = {"datetime": [], "convergence": []}


    def update_convergence_dict(self, current_time: datetime.datetime, convergence:bool):
        """ Updates convergence dict. """
        self.convergence_dict['datetime'].append(current_time)
        self.convergence_dict['convergence'].append(convergence)

    def export_convergence(self, export_path:str = './convergence.csv'):
        """ Export the convergence."""
        
        export_df = pd.DataFrame(self.convergence_dict)
        export_df.to_csv(export_path)


    def simulate(self, subject: Union[observer.MetricsSubject, None] = None):
        """ Loops through all the simulation timesteps """

        while self.current_time <= self.simulation_end_time:
            
            convergence = self.opendss_instance.solve()
            self.update_convergence_dict(self.current_time, convergence)

            if subject:
                subject.notify(self.opendss_instance.dss_instance)
            
            self.current_time += datetime.timedelta(minutes=self.simulation_timestep_min)
            logger.info(f"Simulation finished for {self.current_time} >> {convergence}")


    

        






""" Module for managing deployment strategies."""

import abc
from typing import List
import random

from emerge.scenarios import data_model


class SelectionStrategy(abc.ABC):
    """Abstract class for selection strategy.

    Idea is to take list of customer models and define
    a order in which they are to be selected for placing
    pvs.
    """

    @abc.abstractmethod
    def return_selection_order(
        self, list_of_customers: List[data_model.CustomerModel]
    )-> List[data_model.CustomerModel]:
        """Abstract method for returning the selection order.
        
        Args:
            list_of_customers (List[data_model.CustomerModel]): List
                of `CustomerModel` objects.

        Returns:
            List[data_model.CustomerModel]: List
                of `CustomerModel` objects.
        """


class RandomSelectionStrategy(SelectionStrategy):
    """Implements random selection strategy."""

    def return_selection_order(
        self, list_of_customers: List[data_model.CustomerModel]
    ):
        """ Refer to base class for more details."""
        random.shuffle(list_of_customers)
        return list_of_customers


class CloseSelectionStrategy(SelectionStrategy):
    """Implements close to feeder selection strategy."""

    def return_selection_order(
        self, list_of_customers: List[data_model.CustomerModel]
    ):
        """ Refer to base class for more details."""
        return sorted(list_of_customers, key=lambda d: d.distance)


class FarSelectionStrategy(SelectionStrategy):
    """Implements far from feeder selection strategy."""

    def return_selection_order(
        self, list_of_customers: List[data_model.CustomerModel]
    ):
        """ Refer to base class for more details."""
        return sorted(list_of_customers, key=lambda d: d.distance, reverse=True)


SELECTION_STRATEGY_MAPPER = {
    data_model.SelectionStrategyEnum.random_allocation: RandomSelectionStrategy,
    data_model.SelectionStrategyEnum.far_allocation: FarSelectionStrategy,
    data_model.SelectionStrategyEnum.near_allocation: CloseSelectionStrategy,
}


""" This module contains function for exposing scenarios generation as CLI."""

import click
import json

from emerge.scenarios import (
    data_model,
    scenario,
    opendss_writer,
)
from emerge.utils import dss_util
from emerge.simulator import opendss


@click.command()
@click.option(
    "-c",
    "--config",
    help="Path to config file for generating scenarios",
)
def generate_scenarios(config):
    """Function to create PV deloyment scenarios."""
    # pylint: disable=no-member
    with open(config, "r", encoding="utf-8") as file:
        config_dict = json.load(file)
    
    config_data = data_model.DERScenarioConfigModel.model_validate(
        config_dict)
    
    for der_scen in config_data.der_scenario:

        simulator = opendss.OpenDSSSimulator(config_data.master_file)
        list_of_customers = dss_util.get_list_of_customer_models(simulator.dss_instance, 1, 
                                                                 cust_type=config_data.opendss_attr)
        mapper_object = dss_util.get_load_mapper_objects(simulator.dss_instance)

        derscenarios = scenario.create_der_scenarios(
            list_of_customers, 
            data_model.ScenarioBaseConfig(
                **config_data.model_dump(include=['pct_resolution', 
                                                  'num_of_penetration', 'max_num_of_samples']) 
            ),
            der_config=der_scen
        )
        writer_object = opendss_writer.OpenDSSPVScenarioWriter(
            derscenarios, config_data.output_folder
        )
        writer_object.write(mapper_object,
            file_name=der_scen.file_name         
        )



""" Module for managing computation of system level metrics. """
import copy

import opendssdirect as dss
import networkx as nx

from emerge.metrics import observer
from emerge.simulator import powerflow_results
from emerge.utils import dss_util
from emerge.metrics import data_model
from emerge.network import asset_metrics
import pandas as pd


class TimeseriesTotalLoss(observer.MetricObserver):
    """ Class for computing total loss.
    
    Attributes:
        active_power (float): Time series active power
        reactive_power (float): Time series reactive power
    """

    def __init__(self):

        self.active_power = []
        self.reactive_power = []

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """
        timestep = dss_instance.Solution.StepSize()/(3600)
        sub_losses = dss_instance.Circuit.Losses()
      
        self.active_power.append((sub_losses[0])*timestep/1000)
        self.reactive_power.append((sub_losses[1])*timestep/1000)

    def get_metric(self):
        """ Refer to base class for more details. """
        return {
            "active_power": self.active_power,
            "reactive_power": self.reactive_power
        }

class TimeseriesTotalPower(observer.MetricObserver):
    """ Class for timeseries total power.
    
    Attributes:
        active_power (float): Time series active power
        reactive_power (float): Time series reactive power
    """

    def __init__(self):

        self.active_power = []
        self.reactive_power = []

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """
        timestep = dss_instance.Solution.StepSize()/(3600)
        sub_power = dss_instance.Circuit.TotalPower()
      
        self.active_power.append((-sub_power[0])*timestep/1000)
        self.reactive_power.append((-sub_power[1])*timestep/1000)

    def get_metric(self):
        """ Refer to base class for more details. """
        return {
            "active_power": self.active_power,
            "reactive_power": self.reactive_power
        }

class TimeseriesTotalPVPower(observer.MetricObserver):
    """ Class for computing time series total pv power.
    
    Attributes:
        active_power (float): Time series active power
        reactive_power (float): Time series reactive power
    """

    def __init__(self):

        self.active_power = []
        self.reactive_power = []

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """
        timestep = dss_instance.Solution.StepSize()/(3600)
        pv_df = powerflow_results.get_pv_power_dataframe(dss_instance)
        if not pv_df.empty:
            pv_power = pv_df.sum().to_dict()
            self.active_power.append(pv_power['active_power']*timestep/1000)
            self.reactive_power.append(pv_power['reactive_power']*timestep/1000)
        else:
            self.active_power.append(0)
            self.reactive_power.append(0)
    def get_metric(self):
        """ Refer to base class for more details. """
        return {
            "active_power": self.active_power,
            "reactive_power": self.reactive_power
        }

class TotalLossEnergy(observer.MetricObserver):
    """ Class for computing total loss.
    
    Attributes:
        total_loss (float): Store for total energy
    """

    def __init__(self):

        self.total_loss = {"active_power": 0, "reactive_power": 0}

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """
        timestep = dss_instance.Solution.StepSize()/(3600)
        sub_losses = dss_instance.Circuit.Losses()
      
        self.total_loss['active_power'] += (sub_losses[0])*timestep/1000000
        self.total_loss['reactive_power'] += (sub_losses[1])*timestep/1000000

    def get_metric(self):
        """ Refer to base class for more details. """
        return self.total_loss

class TotalEnergy(observer.MetricObserver):
    """ Class for computing total energy.
    
    Attributes:
        total_energy (float): Store for total energy
    """

    def __init__(self):

        self.total_energy = {"active_power": 0, "reactive_power": 0}

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """
        timestep = dss_instance.Solution.StepSize()/(3600)
        sub_power = dss_instance.Circuit.TotalPower()
      
        self.total_energy['active_power'] += (-sub_power[0])*timestep/1000
        self.total_energy['reactive_power'] += (-sub_power[1])*timestep/1000

    def get_metric(self):
        """ Refer to base class for more details. """
        return self.total_energy

class TotalPVGeneration(observer.MetricObserver):
    """ Class for computing total energy in MWh.
    
    Attributes:
        pv_energy (float): Store for total energy
    """

    def __init__(self):

        self.pv_energy = {"active_power": 0, "reactive_power": 0}

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """
        timestep = dss_instance.Solution.StepSize()/(3600)
        pv_df = powerflow_results.get_pv_power_dataframe(dss_instance)
        if not pv_df.empty:
            pv_power = pv_df.sum().to_dict()
            self.pv_energy['active_power'] += pv_power['active_power']*timestep/1000
            self.pv_energy['reactive_power'] += pv_power['reactive_power']*timestep/1000

    def get_metric(self):
        """ Refer to base class for more details. """
        return self.pv_energy

class SARDI_aggregated(observer.MetricObserver):
    """ Class for computing SARDI aggregated metric.
    
    Attributes:
        loading_limit (ThermalLoadingLimit): Instance of `ThermalLoadingLimit` 
            data model.
        sardi_transformer (float): SARDI_line metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        network (nx.Graph): Networkx graph representing
            distribution network
    """


    def __init__(self, 
        loading_limit: float = 1.0,
        voltage_limit: dict = {
            'overvoltage_threshold': 1.05,
            'undervoltage_threshold': 0.95
            }
        ):
        """ Constructor for `SARDI_line` class. 
        
        Args:
            loading_limit (float): Per unit loading limit
                used for computing SARDI_line metric.
            voltage_limit (dict): Voltage threshold
        """

        self.loading_limit = data_model.ThermalLoadingLimit(threshold=loading_limit)
        self.voltage_limit = data_model.VoltageViolationLimit(**voltage_limit)
        self.sardi_aggregated = 0
        self.counter = 0
    
    def _get_initial_dataset(self, dss_instance: dss):
        
        """ Get initial dataset for computing the metric. """
        self.network = asset_metrics.networkx_from_opendss_model(dss_instance)
        self.load_bus_map = dss_util.get_bus_load_dataframe(dss_instance).set_index("busname")
        self.substation_bus = dss_util.get_source_node(dss_instance)
        self.bus_load_flag_df = dss_util.get_bus_load_flag(dss_instance)

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """

        # Get line loading dataframe
        transformer_loading_df = powerflow_results.get_transloading_dataframe(dss_instance)
        line_loading_df = powerflow_results.get_lineloading_dataframe(dss_instance)
        voltage_df = powerflow_results.get_voltage_dataframe(dss_instance)
        
        if not self.counter:
            self._get_initial_dataset(dss_instance)

        v_filter = voltage_df.loc[self.load_bus_map.index].drop_duplicates()
        ov_flags = v_filter['voltage(pu)']>self.voltage_limit.overvoltage_threshold
        uv_flags = (v_filter['voltage(pu)']<self.voltage_limit.undervoltage_threshold)&(v_filter['voltage(pu)']>0)
        overvoltage_v = v_filter[ov_flags]
        undervoltage_v = v_filter[uv_flags]
        merged_load_buses = list(set(pd.concat([overvoltage_v, undervoltage_v]).index))
        
        self.network_copy = copy.deepcopy(self.network)
        overloaded_trans = transformer_loading_df[transformer_loading_df['loading(pu)']> self.loading_limit.threshold]
        overloaded_lines = line_loading_df[line_loading_df['loading(pu)']> self.loading_limit.threshold]

        if not all([overloaded_trans.empty, overloaded_lines.empty]) :
            
            xfmrs = [] if overloaded_trans.empty else list(overloaded_trans.index)
            lines = [] if overloaded_lines.empty else list(overloaded_lines.index)

            edge_segments = xfmrs + lines
            edge_to_be_removed = []
            
            for edge in self.network_copy.edges():
                edge_data = self.network_copy.get_edge_data(*edge)
                if edge_data and 'name' in edge_data and edge_data['name'] in edge_segments:
                    edge_to_be_removed.append(edge)

            self.network_copy.remove_edges_from(edge_to_be_removed)
            connected_buses = nx.node_connected_component(self.network_copy, self.substation_bus)
            impacted_buses = self.bus_load_flag_df.loc[self.bus_load_flag_df.index.difference(connected_buses)]
            impacted_load_buses = set(impacted_buses[impacted_buses['is_load']==1].index)
            total_impacted_load_buses = impacted_load_buses.union(merged_load_buses)

        else:
            total_impacted_load_buses = merged_load_buses
        
        
        total_load = dss_instance.Loads.Count()
        affected_loads = list(set(self.load_bus_map.loc[list(total_impacted_load_buses)]["loadname"]))
        self.sardi_aggregated += len(affected_loads)*100/total_load
      
        self.counter +=1

    def get_metric(self):
        """ Refer to base class for more details. """
        return {
            "sardi_aggregated": self.sardi_aggregated/self.counter if self.counter >0 else 0
        }


class SARDI_transformer(observer.MetricObserver):
    """ Class for computing SARDI transformer metric.
    
    Attributes:
        loading_limit (ThermalLoadingLimit): Instance of `ThermalLoadingLimit` 
            data model.
        sardi_transformer (float): SARDI_line metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        network (nx.Graph): Networkx graph representing
            distribution network
    """


    def __init__(self, 
        loading_limit: float = 1.0):
        """ Constructor for `SARDI_line` class. 
        
        Args:
            loading_limit (float): Per unit loading limit
                used for computing SARDI_line metric.
        """

        self.loading_limit = data_model.ThermalLoadingLimit(threshold=loading_limit)
        self.sardi_transformer = 0
        self.counter = 0
    
    def _get_initial_dataset(self, dss_instance: dss):
        
        """ Get initial dataset for computing the metric. """
        self.network = asset_metrics.networkx_from_opendss_model(dss_instance)
        self.substation_bus = dss_util.get_source_node(dss_instance)
        self.bus_load_flag_df = dss_util.get_bus_load_flag(dss_instance)
        self.load_bus_map = dss_util.get_bus_load_dataframe(dss_instance).set_index("busname")

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """

        # Get line loading dataframe
        transformer_loading_df = powerflow_results.get_transloading_dataframe(dss_instance)
        
        if not self.counter:
            self._get_initial_dataset(dss_instance)
        
        self.network_copy = copy.deepcopy(self.network)
        overloaded_trans = transformer_loading_df[transformer_loading_df['loading(pu)']> self.loading_limit.threshold]
        
        if not overloaded_trans.empty:
            xfmrs = list(overloaded_trans.index)
            edge_to_be_removed = []
            
            for edge in self.network_copy.edges():
                edge_data = self.network_copy.get_edge_data(*edge)
                if edge_data and 'name' in edge_data and edge_data['name'] in xfmrs:
                    edge_to_be_removed.append(edge)

            self.network_copy.remove_edges_from(edge_to_be_removed)
            connected_buses = nx.node_connected_component(self.network_copy, self.substation_bus)
            impacted_buses = self.bus_load_flag_df.loc[self.bus_load_flag_df.index.difference(connected_buses)]
            impacted_load_buses = impacted_buses[impacted_buses['is_load']==1].index
            affected_loads = list(set(self.load_bus_map.loc[impacted_load_buses]["loadname"]))

            total_load = dss_instance.Loads.Count()
            self.sardi_transformer += len(affected_loads)*100/total_load

        self.counter +=1

    def get_metric(self):
        """ Refer to base class for more details. """
        return {
            "sardi_transformer": self.sardi_transformer/self.counter if self.counter >0 else 0
        }


class SARDI_line(observer.MetricObserver):
    """ Class for computing SARDI line metric.
    
    Attributes:
        loading_limit (ThermalLoadingLimit): Instance of `ThermalLoadingLimit` 
            data model.
        sardi_line (float): SARDI_line metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        network (nx.Graph): Networkx graph representing
            distribution network
    """


    def __init__(self, 
        loading_limit: float = 1.0):
        """ Constructor for `SARDI_line` class. 
        
        Args:
            loading_limit (float): Per unit loading limit
                used for computing SARDI_line metric.
        """

        self.loading_limit = data_model.ThermalLoadingLimit(threshold=loading_limit)
        self.sardi_line = 0
        self.counter = 0
    
    def _get_initial_dataset(self, dss_instance: dss):
        
        """ Get initial dataset for computing the metric. """
        self.network = asset_metrics.networkx_from_opendss_model(dss_instance)
        self.substation_bus = dss_util.get_source_node(dss_instance)
        self.bus_load_flag_df = dss_util.get_bus_load_flag(dss_instance)
        self.load_bus_map = dss_util.get_bus_load_dataframe(dss_instance).set_index("busname")

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """

        # Get line loading dataframe
        line_loading_df = powerflow_results.get_lineloading_dataframe(dss_instance)
        
        if not self.counter:
            self._get_initial_dataset(dss_instance)
        
        self.network_copy = copy.deepcopy(self.network)
        overloaded_lines = line_loading_df[line_loading_df['loading(pu)']> self.loading_limit.threshold]
        
        if not overloaded_lines.empty:
            line_segments = list(overloaded_lines.index)
            edge_to_be_removed = []
            
            for edge in self.network_copy.edges():
                edge_data = self.network_copy.get_edge_data(*edge)
                if edge_data and 'name' in edge_data and edge_data['name'] in line_segments:
                    edge_to_be_removed.append(edge)

            self.network_copy.remove_edges_from(edge_to_be_removed)
            connected_buses = nx.node_connected_component(self.network_copy, self.substation_bus)
            impacted_buses = self.bus_load_flag_df.loc[self.bus_load_flag_df.index.difference(connected_buses)]
            impacted_load_buses = impacted_buses[impacted_buses['is_load']==1].index
            affected_loads = set(list(self.load_bus_map.loc[impacted_load_buses]["loadname"]))
            total_load = dss_instance.Loads.Count()
            self.sardi_line += len(affected_loads)*100/total_load

        self.counter +=1

    def get_metric(self):
        """ Refer to base class for more details. """
        return {
            "sardi_line": self.sardi_line/self.counter if self.counter >0 else 0
        }


class SARDI_voltage(observer.MetricObserver):
    """ Class for managing the computation of SARDI voltage metric.
    
    Attributes:
        upper_threshold (float): Voltage upper threshold
        lower_threshold (float): Voltage lower threshold
        sardi_voltage (float): SARDI voltage metric
        counter (int): Counter for keeping track how number
            of times compute function is called
        load_bus_map (pd.DataFrame): Dataframe containing mapping
            between bus and load
    """

    def __init__(
        self,
        upper_threshold: float = 1.05,
        lower_threshold: float = 0.95
    ):
        """ Constructor for SARDI_voltage metric.
        
        Args:
            upper_threshold (float): Voltage upper threshold
            lower_threshold (float): Voltage lower threshold
        """
        
        self.upper = upper_threshold
        self.lower = lower_threshold
        self.sardi_voltage = 0
        self.counter = 0

    def compute(self, dss_instance:dss):
        """ Refer to base class for more details. """

        # Get voltage dataframe and load bus mapper
        voltage_df = powerflow_results.get_voltage_dataframe(dss_instance)
        
        if not hasattr(self, 'load_bus_map'):
            self.load_bus_map = dss_util.get_bus_load_dataframe(dss_instance).set_index("busname")
        
        # Filter voltages for load buses only 
        # and count the number of overvoltages and undervoltages
        v_filter = voltage_df.loc[self.load_bus_map.index].drop_duplicates()
        overvoltage_df = v_filter[v_filter['voltage(pu)']>self.upper]
        undervoltage_df = v_filter[(v_filter['voltage(pu)']<self.lower)&(v_filter['voltage(pu)']>0)]

        impacted_buses = list(set(list(overvoltage_df.index) + list(undervoltage_df.index)))
        affected_loads = list(set(self.load_bus_map.loc[impacted_buses]["loadname"]))

        self.sardi_voltage += len(affected_loads)*100/len(self.load_bus_map)
        self.counter +=1

    def get_metric(self):
        """ Refer to base class for more details. """
        return {
            "sardi_voltage": self.sardi_voltage/self.counter if self.counter >0 else 0
        }





""" Module for managing export of scenarions in opendss format. """

from typing import List
from pathlib import Path

from emerge.scenarios import data_model

class OpenDSSPVScenarioWriter:
    """Writer class for exporting scenario in opendss format.

    Attributes:
        scenarios (List[data_model.DistPVScenarioModel]): List of pv
            scenarios
        output_path (str): Output path for writing the scenarios.
    """

    def __init__(
        self, scenarios: List[data_model.DistDERScenarioModel], 
        output_path: str
    ):
        """Constructor for `OpenDSSPVScenarioWriter` class.

        Args:
            scenarios (List[data_model.DistPVScenarioModel]): List of pv
                scenarios
            output_path (str): Output path for writing the scenarios.
        """

        self.scenarios = scenarios
        self.output_path = Path(output_path)
        self.output_path.mkdir(exist_ok=True, parents=True)

    def write(self, load_mapper_model: List[data_model.LoadMetadataModel],
               file_name: str )-> None:
        """Method for writing the scenarios.
        
        Args:
            load_mapper_model (List[data_model.LoadMetadataModel]): List of load 
                metadata model.
            file_name (str): OpenDSS file name: e.g new_pvs.dss
        """
        for scenario in self.scenarios:
            
            scenario_folder = self.output_path / scenario.name

            if not scenario_folder.exists():
                scenario_folder.mkdir(parents=True, exist_ok=True)

            der_models = []
            der_models.append(
                f"! DER Scenario for {scenario.penetration} kW total size, Sample {scenario.sample_id} \n"
            )
            der_models.append(
                "==============================================DER SCENARIO FILE====================================\n\n"
            )

            for der in scenario.ders:
                mapper = next(
                    filter(lambda d: d.name == der.customer.name, load_mapper_model)
                )
                
                if der.der_type == data_model.DERType.solar:
                    der_models.append(
                        f"new pvsystem.{der.name.replace('.', '_')} phases={mapper.num_phase}"
                        f" bus1={mapper.bus} kv={mapper.kv} irradiance=1 pmpp={der.kw} kva={der.kw}"
                        f" conn=wye %cutin=0.1 %cutout=0.1 vmaxpu=1.2 yearly={der.profile}\n" 
                    )
                elif der.der_type == data_model.DERType.load:
                    if der.kw !=0 :
                        der_models.append(
                            f"new load.{der.name.replace('.', '_')} phases={mapper.num_phase}"
                            f" bus1={mapper.bus} kv={mapper.kv} kvar={0} kva={der.kw}"
                            f" yearly={der.profile}\n" 
                        )
                else:
                    raise NotImplementedError(f"{der.der_type} has not been implementes.")
                

            der_path = scenario_folder / file_name
            with open(der_path, "w", encoding='utf-8') as file:
                file.writelines(der_models)



""" This module implements commad line interface for exporting 
schemas and handling opt in vscode update."""

# standard imports
from pathlib import Path
from typing import List

# third-party imports
import click
from pydantic import BaseModel
import pydantic

# internal imports
from emerge.scenarios import data_model
from emerge.utils import util

class SchemaItemModel(BaseModel):
    name: str 
    model: pydantic._internal._model_construction.ModelMetaclass

    class Config:
        arbitrary_types_allowed=True

class SchemaManager:
    def __init__(self, schema_folder: str =".vscode"):
        self.schema_folder = Path(schema_folder)
        self.schemas: List[SchemaItemModel] = []

    def add_schema(self, name: str, model: BaseModel):
        self.schemas.append(SchemaItemModel(name=name, model=model))

    def generate_and_save_schemas(self):
        
        if not self.schema_folder.exists():
            self.schema_folder.mkdir()

        for schema in self.schemas:
            json_schema = schema.model.model_json_schema()
            schema_file = self.schema_folder / f"{schema.name}.json"
            util.write_file(json_schema, schema_file)

    def configure_vscode_settings(self):
        vscode_settings_file = self.schema_folder / "settings.json"
        
        if not vscode_settings_file.exists():
            util.write_file({}, vscode_settings_file)

        settings = util.read_file(vscode_settings_file, use_json5=True)
        vscode_key = "json.schemas"

        if vscode_key not in settings:
            settings[vscode_key] = []

        for schema in self.schemas:
            updated_in_place = False
            schema_file = self.schema_folder / f"{schema.name}.json"

            for item in settings[vscode_key]:
                if schema.name in item.get("url", ''):
                    item["url"] = str(schema_file)
                    updated_in_place = True

            if not updated_in_place:
                settings[vscode_key].append({
                    "fileMatch": ["*.json"],
                    "url": str(schema_file)
                })

        util.write_file(settings, vscode_settings_file)

@click.command()
@click.option(
    "-vc",
    "--vscode",
    default=False,
    show_default=True,
    help="""Update JSON schemas in vscode settings. Note will create .vscode folder 
    if not present in the current directory."""
)
def create_schemas(vscode: bool):
    """ Function to handle the JSON schemas for emerge package."""

    schema_manager = SchemaManager()
    schema_manager.add_schema("emerge_scenario_schema", 
                              data_model.DERScenarioConfigModel)
    schema_manager.generate_and_save_schemas()
    if vscode:
        schema_manager.configure_vscode_settings()


""" Module for generating DER deployment scenarios.

A distributed energy resource (DER) deployment scenario is a hypothetical scenario defined on top
of base distribution system model that uses some rules/forecast/expertise
to come up with number, size and location of der units to be installed.
"""
# standard imports
from typing import List


# internal imports
from emerge.scenarios import data_model
from emerge.scenarios import selection_strategy, sizing_strategy
from emerge.utils import dss_util
from emerge.simulator import opendss
from emerge.scenarios import opendss_writer

def _get_sizing_func_input(item: data_model._DERScenarioInput):
    return {
        data_model.CapacityStrategyEnum.solar_energy_based: item.energy_sizing_input,
        data_model.CapacityStrategyEnum.peak_multiplier: item.peakmult_sizing_input,
        data_model.CapacityStrategyEnum.fixed_sizing: item.fixed_sizing_input
    }.get(item.sizing_strategy)



def _add_other_ders(
    ders_list: List[data_model.BasicDERModel],
    strategy: sizing_strategy.SizingStrategy,
    der_type: data_model.DERType,
    der_tag: str = ''
)-> List[data_model.BasicDERModel]:
    """Add other ders on top of existing der for a customer.

    Args:
        ders_list (List[data_model.BasicDERModel]): List of already existing DERs
        strategy (sizing_strategy.SizingStrategy): Strategy for coming up with the DER size
        der_type (data_model.DERType): DERType to be added
    """

    der_models = []

    for der in ders_list:
        if der.der_type == der_type:
            # pylint: disable=broad-exception-raised
            raise ValueError(f'DER type can not be same when adding other ders! {der} >> {der_type}')
        
        der_cap, profile = strategy.return_kw_and_profile(der.customer)

        if der_cap and profile:
            der_models.append(
                data_model.BasicDERModel(
                        name=der.name + f'_{der_type}_{der_tag}',
                        kw=der_cap,
                        customer=der.customer,
                        der_type=der_type,
                        der_tag=der_tag,
                        profile=profile
                    )
            )
    return der_models


def _get_ders(
    customers: List[data_model.CustomerModel],
    target_kw: float,
    strategy: sizing_strategy.SizingStrategy,
    der_type: data_model.DERType,
    der_tag: str = ''
) -> List[data_model.BasicDERModel]:
    """Creates a list of der systems for a target kw.

    Args:
        customers (List[data_model.CustomerModel]): List of customers models
        target_kw (float): Target kW
        strategy (sizing_strategy.SizingStrategy): Strategy for coming up with the DER size
        der_type (data_model.DERType): DERType to be added
    """

    if not customers:
        return []

    der_models = []
    for customer in customers:
        if target_kw <= 0:
            break

        der_capacity, profile = strategy.return_kw_and_profile(customer)

        if der_capacity and profile:
            der_capacity = min(der_capacity, target_kw)

            target_kw -= der_capacity
            der_models.append(
                data_model.BasicDERModel(
                    name=customer.name + "_der",
                    kw=der_capacity,
                    customer=customer,
                    der_type=der_type,
                    der_tag=der_tag,
                    profile=profile
                )
            )

    return der_models


# pylint: disable=too-many-locals
def create_der_scenarios(
    list_of_customers: List[data_model.CustomerModel],
    scenario_config: data_model.ScenarioBaseConfig,
    der_config: data_model.DERScenarioInputModel
) -> List[data_model.DistDERScenarioModel]:
    """Function for creating distributed DER scenarios
    for distribution system.

    Args:
        list_of_loads (List[data_model.CustomerModel]): List of customer for
                scenario generation.
        scenario_config (data_model.ScenarioBaseConfig): Basic configuration data for
                generating scenarios
        der_config List[data_model.DERScenarioInputModel]: DER config.
    Returns:
        List[data_model.DistDERScenarioModel]: List of `DistDERScenarioModel`
            models.
    """

    select_strategy = selection_strategy.SELECTION_STRATEGY_MAPPER[
        der_config.selection_strategy]()
    
    scenarios: List[data_model.DistDERScenarioModel] = []

    # Get total load
    total_load_kw = sum(load.kw for load in list_of_customers)

    # loop over all the samples and pv penetrations
    for sample_id in range(scenario_config.max_num_of_samples):
        # Idea is to keep pv systems installed in previous penetrations
        past_ders: List[data_model.BasicDERModel] = []
        past_capacity = 0

        for der_penetration_id in range(1, scenario_config.num_of_penetration + 1):
            
            der_capacity = (
                der_penetration_id * scenario_config.pct_resolution / 100
            ) * total_load_kw

            previous_loads = list(set([der.customer.name for der in past_ders]))
            loads_without_der = [
                load for load in list_of_customers if load.name not in previous_loads
            ]

            # Returns order in which load is to be selected to add pv
            loads = select_strategy.return_selection_order(
                list_of_customers=loads_without_der
            )
            
            sizing_strategy_object = sizing_strategy.SIZING_STRATEGY_MAPPER.get(
                der_config.sizing_strategy
            )(_get_sizing_func_input(der_config))


            main_der_models = _get_ders(loads, der_capacity - past_capacity, 
                                   sizing_strategy_object, der_config.der_type, der_config.der_tag)
            
            other_der_models = []
            for other_der in der_config.other_ders:
                sizing_strategy_object = sizing_strategy.SIZING_STRATEGY_MAPPER.get(
                    other_der.sizing_strategy
                )(_get_sizing_func_input(other_der))
                other_der_models += _add_other_ders(main_der_models, 
                                   sizing_strategy_object, other_der.der_type, other_der.der_tag)
            
            scenario = data_model.DistDERScenarioModel(
                name=f"scenario_{sample_id}_"
                f"{int(der_penetration_id*scenario_config.pct_resolution)}",
                sample_id=sample_id,
                penetration=round(der_capacity, 3),
                ders=past_ders + main_der_models + other_der_models,
            )

            past_capacity = der_capacity
            past_ders = past_ders + main_der_models + other_der_models

            scenarios.append(scenario)

    return scenarios

def generate_scenarios(config_data: data_model.DERScenarioConfigModel )-> None:
    """ Generate scenarios by taking high level configuration and writes opendss files.
    
    Args:
        list_of_loads (data_model.DERScenarioConfigModel): Configuration model for defining
            how to generate DER scenarios.
    """

    for der_scen in config_data.der_scenario:

        simulator = opendss.OpenDSSSimulator(config_data.master_file)
        list_of_customers = dss_util.get_list_of_customer_models(simulator.dss_instance, 1, 
                                                                 cust_type=config_data.opendss_attr)
        mapper_object = dss_util.get_load_mapper_objects(simulator.dss_instance)

        derscenarios = create_der_scenarios(
            list_of_customers, data_model.ScenarioBaseConfig(
                **config_data.model_dump(include=['pct_resolution', 
                                                  'num_of_penetration', 'max_num_of_samples']) 
            ),
            der_config=der_scen
        )
        writer_object = opendss_writer.OpenDSSPVScenarioWriter(
            derscenarios, config_data.output_folder
        )
        writer_object.write(mapper_object,
            file_name=der_scen.file_name         
        )



""" Module for computing time series metrics for multi scenario simulation."""
# standard imports
import datetime
from pathlib import Path
import multiprocessing

# third-party imports
import click

# internal imports
from emerge.metrics import system_metrics
from emerge.metrics import observer
from emerge.metrics import node_metrics
from emerge.simulator import simulation_manager


def _run_timeseries_sim(input):

    date_format = "%Y-%m-%d %H:%M:%S"
    manager = simulation_manager.OpenDSSSimulationManager(
        input["master_file"],
        datetime.datetime.strptime(input["simulation_start"], date_format),
        datetime.datetime.strptime(input["profile_start"], date_format),
        datetime.datetime.strptime(input["simulation_end"], date_format),
        input["simulation_resolution"],
    )
    manager.opendss_instance.execute_dss_command(f"Redirect {input['pv_file']}")
    manager.opendss_instance.set_max_iteration(200)
    subject = observer.MetricsSubject()

    sardi_voltage_observer = system_metrics.SARDI_voltage(
        input["overvoltage_threshold"], input["undervoltage_threshold"]
    )
    sardi_line_observer = system_metrics.SARDI_line(input["thermal_threshold"])
    sardi_xfmr_observer = system_metrics.SARDI_transformer(
        input["thermal_threshold"]
    )
    sardi_aggregated_observer = system_metrics.SARDI_aggregated(
        loading_limit=input["thermal_threshold"],
        voltage_limit={
            "overvoltage_threshold": input["overvoltage_threshold"],
            "undervoltage_threshold": input["undervoltage_threshold"],
        },
    )
    nvri_observer = node_metrics.NVRI(
        input["overvoltage_threshold"], input["undervoltage_threshold"]
    )
    llri_observer = node_metrics.LLRI(input["thermal_threshold"])
    tlri_observer = node_metrics.TLRI(input["thermal_threshold"])
    total_energy_observer = system_metrics.TotalEnergy()
    total_pv_energy_observer = system_metrics.TotalPVGeneration()
    timeseries_total_power_observer = system_metrics.TimeseriesTotalPower()
    timeseries_total_pv_power_observer = system_metrics.TimeseriesTotalPVPower()
    total_loss_observer = system_metrics.TotalLossEnergy()
    timeseries_loss_observer = system_metrics.TimeseriesTotalLoss()

    observers_ = [
        sardi_voltage_observer,
        sardi_line_observer,
        sardi_xfmr_observer,
        sardi_aggregated_observer,
        nvri_observer,
        llri_observer,
        tlri_observer,
        total_energy_observer,
        total_pv_energy_observer,
        timeseries_total_power_observer,
        timeseries_total_pv_power_observer,
        total_loss_observer,
        timeseries_loss_observer
    ]
    for observer_ in observers_:
        subject.attach(observer_)

    manager.opendss_instance.execute_dss_command(f"batchedit pvsystem..* yearly={input['solar_profile']}")

    if input['voltvar']:
        num_points = len(input['voltvar_yarray'].split(','))
        curve_command = f"new xycurve.vvar_curve npts={num_points} yarray={input['voltvar_yarray']} xarray={input['voltvar_xarray']}"
        manager.opendss_instance.execute_dss_command(curve_command)
        
        inverter_command = f"new invcontrol.inv_controller_ mode=VOLTVAR voltage_curvex_ref=rated vvc_curve1=vvar_curve eventlog=yes RefReactivePower=VARMAX_VARS"
        manager.opendss_instance.execute_dss_command(inverter_command)

        manager.opendss_instance.execute_dss_command("batchedit pvsystem..* wattpriority=yes")
        if not input['nighttime']:
            manager.opendss_instance.execute_dss_command("batchedit pvsystem..* varfollowinverter=yes")

    manager.simulate(subject)
    manager.export_convergence(input["convergence_report"])
    observer.export_tinydb_json(observers_, input["output_json"])


@click.command()
@click.option(
    "-m",
    "--master-file",
    help="Path to master dss file",
)
@click.option(
    "-sf",
    "--scenario-folder",
    help="Path to a scenario folder",
)
@click.option(
    "-pn",
    "--profile-name",
    help="Profile name for solar.",
)
@click.option(
    "-nc",
    "--number-of-cores",
    default=1,
    show_default=True,
    help="Number of cores to be used in parallel",
)
@click.option(
    "-ss",
    "--simulation-start",
    default="2022-1-1 00:00:00",
    show_default=True,
    help="Simulation start time.",
)
@click.option(
    "-ps",
    "--profile-start",
    default="2022-1-1 00:00:00",
    show_default=True,
    help="Time series profile start time.",
)
@click.option(
    "-se",
    "--simulation-end",
    default="2022-1-2 00:00:00",
    show_default=True,
    help="Simulation end time.",
)
@click.option(
    "-r",
    "--simulation-resolution",
    default=60,
    show_default=True,
    help="Simulation time resolution in minutes.",
)
@click.option(
    "-ot",
    "--overvoltage-threshold",
    default=1.05,
    show_default=True,
    help="Overvoltage threshold.",
)
@click.option(
    "-ut",
    "--undervoltage-threshold",
    default=0.95,
    show_default=True,
    help="Undervoltage threshold.",
)
@click.option(
    "-tt",
    "--thermal-threshold",
    default=0.95,
    show_default=True,
    help="Thermal laoding threshold.",
)
@click.option(
    "-vvar",
    "--voltvar",
    default=False,
    show_default=True,
    help="Voltvar smart inverter settings."
)
@click.option(
    "-qx",
    "--voltvar-xarray",
    default="[0.7,0.92,0.967,1.033,1.07,1.3]",
    show_default=True,
    help="X array for voltvar curve"
)
@click.option(
    "-qy",
    "--voltvar-yarray",
    default="[0.44,0.44,0,0,-0.44, -0.44]",
    show_default=True,
    help="Y array for voltvar curve"
)
@click.option(
    "-nt",
    "--nighttime",
    default=True,
    show_default=True,
    help="Inverter night time volt var setting."
)
@click.option(
    "-o",
    "--output-folder",
    help="Path to a scenario folder",
)
def compute_multiscenario_time_series_metrics(
    master_file,
    scenario_folder,
    profile_name,
    number_of_cores,
    simulation_start,
    profile_start,
    simulation_end,
    simulation_resolution,
    overvoltage_threshold,
    undervoltage_threshold,
    thermal_threshold,
    voltvar,
    voltvar_xarray,
    voltvar_yarray,
    nighttime,
    output_folder,
):
    """Run multiscenario time series simulation and compute
    time series metrics."""

    scenario_folder = Path(scenario_folder)
    output_folder = Path(output_folder)

    output_folder.mkdir(exist_ok=True)

    timeseries_input = []

    for folder_path in scenario_folder.iterdir():
        pvsystem_file = folder_path / "PVSystems.dss"
        json_path = output_folder / (folder_path.name + ".json")
        convergence_file = output_folder / (folder_path.name + "_convergence.csv")
        timeseries_input.append(
            {
                "master_file": master_file,
                "simulation_start": simulation_start,
                "solar_profile": profile_name,
                "profile_start": profile_start,
                "simulation_end": simulation_end,
                "simulation_resolution": simulation_resolution,
                "pv_file": pvsystem_file,
                "overvoltage_threshold": overvoltage_threshold,
                "undervoltage_threshold": undervoltage_threshold,
                "thermal_threshold": thermal_threshold,
                "output_json": json_path,
                "convergence_report": convergence_file,
                "voltvar": voltvar,
                "nighttime": nighttime,
                "voltvar_xarray": voltvar_xarray,
                "voltvar_yarray": voltvar_yarray
            }
        )

    with multiprocessing.Pool(number_of_cores) as p:
        p.map(_run_timeseries_sim, timeseries_input)



""" Package setup file. """

from setuptools import setup, find_packages

with open("README.md","r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements = f.read().splitlines()

setup(
    name='NREL-emerge',
    long_description=long_description,
    long_description_content_type="text/markdown",
    version='1.0.0',
    description='Emerging technologies Management and Risk evaluation on distributions Grid Evolution',
    author='Kapil Duwadi',
    author_email='kapil.duwadi@nrel.gov',
    packages=find_packages(),
    url="https://github.com/NREL/EMeRGE",
    keywords="Distribution System DER technologies management risk impact analysis",
    install_requires=requirements,
    # package_dir={"emerge": "emerge"}, 
    entry_points={
        "console_scripts": [
            "emerge=emerge.cli.cli:cli"
        ],
    },
    python_requires=">=3.10",  
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ]
)


""" Module for computing timeseries voltage stats. """

# standard imports
from typing import Dict, List

# third-party imports
import opendssdirect as dss
import polars 

# internal imports
from emerge.metrics import observer
from emerge.simulator import powerflow_results


def get_voltage_df(dss_instance:dss):
    return polars.from_pandas(
            powerflow_results.get_voltage_dataframe(
            dss_instance
        ))

class NodeVoltageStats(observer.MetricObserver):
    """ Class for computing node voltage statistical metrics. """

    def __init__(self):
        
        self.metrics = {
            metric: [] for metric in [
                "min", "max", "median", "mean", "quantile_0.1",
                "quantile_0.25", "quantile_0.75", "quantile_0.90"
            ]
        }

    def compute(self, dss_instance:dss) -> None:
        df = get_voltage_df(dss_instance)
        for metric in self.metrics:
            
            if '_' not in metric:
                self.metrics[metric].append(
                    eval(f"df.{metric}()['voltage(pu)'][0]"))
            else:
                param = float(metric.split('_')[1])
                self.metrics[metric].append(
                    eval(f"df.{metric.split('_')[0]}({param})['voltage(pu)'][0]"))
    

    def get_metric(self) -> Dict:
        return self.metrics

class NodeVoltageBins(observer.MetricObserver):
    """ Class for computing node voltage statistical metrics. """

    def __init__(self, bins: List[float]):
        
        bins = list(set(bins)) 
        bins.sort()

        if len(bins) < 2:
            raise ValueError(f"{bins} should at least have two unique items.")
        
        self.metrics = {
            f'>{bins[id]}__<{el}': 0 for id, el in enumerate(bins[1:])
        }
        self.metrics[f'<{bins[0]}'] = 0
        self.metrics[f'>{bins[-1]}'] = 0


    def compute(self, dss_instance:dss) -> None:
        df = get_voltage_df(dss_instance)
        timestep = dss_instance.Solution.StepSize()/(3600)
        
        for metric in self.metrics:
            
            if '__' in metric:
                low_val = float(metric.split('__')[0].split('>')[1])
                high_val = float(metric.split('__')[1].split('<')[1])

                self.metrics[metric] += len(df.filter((polars.col('voltage(pu)')>low_val)\
                    &(polars.col('voltage(pu)')<high_val)))*timestep

            elif '<' in metric and '__' not in metric:
                low_val = float(metric.split('<')[1])
                self.metrics[metric] += len(df.filter(\
                    (polars.col('voltage(pu)')<low_val)))*timestep

            elif '>' in metric and '__' not in metric:
                high_val = float(metric.split('>')[1])
                self.metrics[metric] += len(df.filter(\
                    (polars.col('voltage(pu)')>high_val)))*timestep


    def get_metric(self) -> Dict:
        return self.metrics




""" Module for computing time series metrics for multi scenario simulation."""

# standard imports
import datetime
from pathlib import Path
from typing import Dict
import concurrent.futures

# third-party imports
import click
import yaml

# internal imports
from emerge.metrics import system_metrics
from emerge.metrics import observer
from emerge.simulator import simulation_manager
from emerge.metrics import node_voltage_stats
from emerge.metrics import line_loading_stats
from emerge.metrics import xfmr_loading_stats


def _get_observers(metrics: Dict):
    observers = {}

    class_mapper = {
        'substation_power': system_metrics.TimeseriesTotalPower,
        'total_pvpower': system_metrics.TimeseriesTotalPVPower,
        'total_powerloss': system_metrics.TimeseriesTotalLoss,
        'node_voltage_stats': node_voltage_stats.NodeVoltageStats,
        'node_voltage_bins': node_voltage_stats.NodeVoltageBins,
        'line_loading_stats': line_loading_stats.LineLoadingStats,
        'line_loading_bins': line_loading_stats.LineLoadingBins,
        'xfmr_loading_stats': xfmr_loading_stats.XfmrLoadingStats,
        'xfmr_loading_bins': xfmr_loading_stats.XfmrLoadingBins,
        'overloaded_lines': line_loading_stats.OverloadedLines,
        'overloaded_transformers': xfmr_loading_stats.OverloadedTransformers,
        'sardi_voltage': system_metrics.SARDI_voltage,
        'sardi_line': system_metrics.SARDI_line,
        'sardi_xfmr': system_metrics.SARDI_transformer,
        'sardi_aggregated': system_metrics.SARDI_aggregated
    }

    for key, subdict in metrics.items():
        if key in class_mapper:
            if 'args' not in subdict:
                observers[key] = class_mapper[key]()
            else:
                observers[key] = class_mapper[key](*subdict['args'])
    return observers

def _run_timeseries_sim(
        master_dss_file_path: str,
        start_time: str, 
        end_time: str, 
        profile_start_time: str,
        resolution_min: float,
        export_path: str,
        time_group: str,
        metrics,
        scenario_base_folder: str,
        scenario_folder_path = None,
):
    date_format = "%Y-%m-%d %H:%M:%S"
    manager = simulation_manager.OpenDSSSimulationManager(
                master_dss_file_path,
                datetime.datetime.strptime(start_time, date_format),
                datetime.datetime.strptime(profile_start_time, date_format),
                datetime.datetime.strptime(end_time, date_format),
                resolution_min,
            )
    manager.opendss_instance.set_max_iteration(200)
    subject = observer.MetricsSubject() 

    if scenario_folder_path:

        for file_path in scenario_folder_path.iterdir():
            manager.opendss_instance.execute_dss_command(f"Redirect {file_path}")


    observers = _get_observers(metrics)
    for _, observer_ in observers.items():
        subject.attach(observer_)

    manager.simulate(subject)

    export_base_path = Path(export_path) / f"{scenario_base_folder}_{time_group}"
    
    if scenario_folder_path:
        export_base_path = export_base_path / scenario_folder_path.name
    
    if not export_base_path.exists():
        export_base_path.mkdir(parents=True)
    
    manager.export_convergence(export_base_path / 'convergence_report.csv')
    observer.export_csv(list(observers.values()), export_base_path)


def _compute_scenario_custom_metrics_wrapper(input):
    _compute_scenario_custom_metrics(*input)

def _compute_scenario_custom_metrics(
    config, 
    master_dss_file,
    scenario_base_folder,
    scenario_folder_path=None, 
):
    

    for time_group, time_dict in config['time_group'].items():
        _run_timeseries_sim(
            master_dss_file,
            time_dict['start_time'],
            time_dict['end_time'],
            config["profile_start"],
            config["resolution_min"],
            config["export_path"],
            time_group,
            config["metrics"],
            scenario_base_folder,
            scenario_folder_path,
        )
        
        
@click.command()
@click.option(
    "-c",
    "--config-file",
    help="Path to config yaml file.",
)
def compute_custom_metrics(
    config_file
):
    """Compute custom metrics after running time series powerflow."""

    with open(config_file, "r") as fp:
        config = yaml.safe_load(fp)


    if config.get('pv_scenarios', None):
       
        for folder_, master_file in config['pv_scenarios']['scenario_folder'].items():
            scen_folder = Path(config['pv_scenarios']['scenario_base_path']) / folder_
            if config['pv_scenarios']['num_core'] > 1:
                
                all_paths = list(scen_folder.iterdir())

                sim_inputs = list(zip(
                    [config]*len(all_paths),
                    [master_file]*len(all_paths),
                    [folder_]*len(all_paths),
                    all_paths
                ))
                with concurrent.futures.ProcessPoolExecutor(
                    max_workers=config['pv_scenarios']['num_core']) as executor:
                    for input, _ in zip(sim_inputs, 
                                        executor.map(_compute_scenario_custom_metrics_wrapper, 
                                        sim_inputs)):
                        print(f'{input} completed ...')
                # with multiprocessing.Pool(config['pv_scenarios']['num_core']) as p:
                #     _ = p.starmap(_compute_scenario_custom_metrics, sim_inputs )
            else:
                for path in scen_folder.iterdir():
                    _compute_scenario_custom_metrics(
                        config,
                        master_file,
                        folder_,
                        path  
                    )
    
    if config.get('base_scenarios', None):
        
        if config['base_scenarios']['num_core'] > 1:
            
            all_master_files = list(config['base_scenarios']['scenario_folder'].values())
            sim_inputs =  list(zip(
                    [config]*len(all_master_files),
                    all_master_files,
                    list(config['base_scenarios']['scenario_folder'].keys()),
                    [None]*len(all_master_files),
                ))
            
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=config['base_scenarios']['num_core']) as executor:
                for input, _ in zip(sim_inputs, executor.map(
                        _compute_scenario_custom_metrics_wrapper, 
                        sim_inputs)):
                    print(f'{input} completed ...')

            # with multiprocessing.Pool(config['base_scenarios']['num_core']) as p:
            #     _ = p.starmap(_compute_scenario_custom_metrics, sim_inputs)
        else:
            for key, path in config['base_scenarios']['scenario_folder'].items():
                _compute_scenario_custom_metrics(config, [], path, key, None, None )



    

    


"""
Command line utility for EMeRGE
"""

# standard imports
from pathlib import Path
import datetime

# third-party imports
import click

# internal imports
from emerge.metrics.snapshot_metrics import compute_snapshot_metrics
from emerge.metrics import system_metrics
from emerge.metrics import observer
from emerge.metrics import node_metrics
from emerge.simulator import simulation_manager
from emerge.network import feeder_geojson
from emerge.simulator import opendss
from emerge.cli.scenario import generate_scenarios
from emerge.cli.multiscenario_metrics import (
    compute_multiscenario_time_series_metrics,
)
from emerge.cli.custom_metrics import compute_custom_metrics
from emerge.cli.schema_generator import create_schemas


@click.command()
@click.option(
    "-m",
    "--master-file",
    help="Path to master dss file",
)
@click.option(
    "-o",
    "--output-folder",
    default="./geojsons",
    show_default=True,
    help="Ouput directory for storing the geojsons",
)
def create_geojsons(master_file, output_folder):
    """Command line function to generate geojsons from opendss model"""
    
    opendss_instance = opendss.OpenDSSSimulator(master_file)
    feeder_geojson.create_feeder_geojson(
        opendss_instance.dss_instance, output_folder
    )


@click.command()
@click.option(
    "-m",
    "--master-file",
    help="Path to master dss file",
)
@click.option(
    "-ss",
    "--simulation-start",
    default="2022-1-1 00:00:00",
    show_default=True,
    help="Simulation start time.",
)
@click.option(
    "-ps",
    "--profile-start",
    default="2022-1-1 00:00:00",
    show_default=True,
    help="Time series profile start time.",
)
@click.option(
    "-se",
    "--simulation-end",
    default="2022-1-2 00:00:00",
    show_default=True,
    help="Simulation end time.",
)
@click.option(
    "-r",
    "--simulation-resolution",
    default=60,
    show_default=True,
    help="Simulation time resolution in minutes.",
)
@click.option(
    "-ot",
    "--overvoltage-threshold",
    default=1.05,
    show_default=True,
    help="Overvoltage threshold.",
)
@click.option(
    "-ut",
    "--undervoltage-threshold",
    default=0.95,
    show_default=True,
    help="Undervoltage threshold.",
)
@click.option(
    "-tt",
    "--thermal-threshold",
    default=0.95,
    show_default=True,
    help="Thermal laoding threshold.",
)
@click.option(
    "-o",
    "--output-json",
    default="db_metric.json",
    show_default=True,
    help="Ouput directory for storing the db",
)
def compute_time_series_metrics(
    master_file,
    simulation_start,
    profile_start,
    simulation_end,
    simulation_resolution,
    overvoltage_threshold,
    undervoltage_threshold,
    thermal_threshold,
    output_json,
):
    """Reads the OpenDSS model and computes various snapshot
    metrics which can be later ingested by dashboard."""

    date_format = "%Y-%m-%d %H:%M:%S"
    manager = simulation_manager.OpenDSSSimulationManager(
        master_file,
        datetime.datetime.strptime(simulation_start, date_format),
        datetime.datetime.strptime(profile_start, date_format),
        datetime.datetime.strptime(simulation_end, date_format),
        simulation_resolution,
    )
    subject = observer.MetricsSubject()

    sardi_voltage_observer = system_metrics.SARDI_voltage(
        overvoltage_threshold, undervoltage_threshold
    )
    sardi_line_observer = system_metrics.SARDI_line(thermal_threshold)
    sardi_xfmr_observer = system_metrics.SARDI_transformer(thermal_threshold)
    sardi_aggregated_observer = system_metrics.SARDI_aggregated(
        loading_limit=thermal_threshold,
        voltage_limit={
            "overvoltage_threshold": overvoltage_threshold,
            "undervoltage_threshold": undervoltage_threshold,
        },
    )
    nvri_observer = node_metrics.NVRI(
        overvoltage_threshold, undervoltage_threshold
    )
    llri_observer = node_metrics.LLRI(thermal_threshold)
    tlri_observer = node_metrics.TLRI(thermal_threshold)

    observers_ = [
        sardi_voltage_observer,
        sardi_line_observer,
        sardi_xfmr_observer,
        sardi_aggregated_observer,
        nvri_observer,
        llri_observer,
        tlri_observer,
    ]
    for observer_ in observers_:
        subject.attach(observer_)

    manager.simulate(subject)
    observer.export_tinydb_json(observers_, output_json)


@click.command()
@click.option(
    "-m",
    "--master-file",
    help="Path to master dss file",
)
@click.option(
    "-o",
    "--output-json",
    default="db.json",
    show_default=True,
    help="Ouput directory for storing the db",
)
def snapshot_metrics(master_file, output_json):
    """Compute snapshot metrics."""
    compute_snapshot_metrics(Path(master_file), output_json)


@click.group()
def cli():
    """Entry point"""

cli.add_command(snapshot_metrics)
cli.add_command(compute_time_series_metrics)
cli.add_command(create_geojsons)
cli.add_command(generate_scenarios)
cli.add_command(compute_multiscenario_time_series_metrics)
cli.add_command(compute_custom_metrics)
cli.add_command(create_schemas)




"""
 Extract base level metrics
"""
# standard imports

# third-party imports
from pydantic import (
        Field, BaseModel)
from typing_extensions import Annotated

class LoadAssetMetrics(BaseModel):
    total_count: Annotated[int, Field(ge=0)] = 0
    max_kw_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    min_kw_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    min_kvar_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    max_kvar_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    total_kw_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    total_kvar_capacity: Annotated[float, Field(ge=0.0)] = 0.0

class PVAssetMetrics(BaseModel):
    total_count: Annotated[int, Field(ge=0)] = 0
    max_kw_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    min_kw_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    total_kw_capacity: Annotated[float, Field(ge=0.0)] = 0.0

class CapacitorAssetMetrics(BaseModel):
    total_count: Annotated[int, Field(ge=0)] = 0
    max_kvar_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    min_kvar_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    total_kvar_capacity: Annotated[float, Field(ge=0.0)] = 0.0

class RegulatorsAssetMetrics(BaseModel):
    total_count: Annotated[int, Field(ge=0)] = 0

class TransformersAssetMetrics(BaseModel):
    total_count: Annotated[int, Field(ge=0)] = 0
    max_kva_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    min_kva_capacity: Annotated[float, Field(ge=0.0)] = 0.0
    total_kva_capacity: Annotated[float, Field(ge=0.0)] = 0.0

class FeederMetrics(BaseModel):
    total_feeder_length_km: Annotated[float, Field(ge=0.0)] = 0.0
    max_primary_feeder_length_km: Annotated[float, Field(ge=0.0)] = 0.0
    max_secondary_feeder_length_km:  Annotated[float, Field(ge=0.0)] = 0.0
    min_secondary_feeder_length_km: Annotated[float, Field(ge=0.0)] = 0.0
    primary_kv_level: Annotated[float, Field(ge=0.0)] = 0.0
    secondary_kv_level: Annotated[float, Field(ge=0.0)] = 0.0
    total_buses: Annotated[int, Field(ge=0)] = 0
    total_line_sections: Annotated[int, Field(ge=0)] = 0


class AssetMetrics(BaseModel):
    loads: LoadAssetMetrics
    pvs: PVAssetMetrics
    capacitors: CapacitorAssetMetrics
    regulators: RegulatorsAssetMetrics
    transformers: TransformersAssetMetrics
    lines: FeederMetrics
    






